import numpy as np
from kerasAC.metrics import *
import keras;
from keras.models import Model
from keras.layers import Input
from keras.layers.core import Dropout, Reshape, Dense, Activation, Flatten
from keras.layers.convolutional import Conv2D, MaxPooling2D
from keras.layers.merge import Add 
from keras.optimizers import Adadelta, SGD, RMSprop;
import keras.losses;
from keras.constraints import maxnorm;
from keras.layers.normalization import BatchNormalization
from keras.regularizers import l1, l2
from keras import backend as K


def l1_block(x):
    tmp=Conv2D(filters=300,kernel_size=(1,9),padding="same")(x)
    tmp=BatchNormalization(axis=-1)(tmp)
    tmp=Activation('relu')(tmp)
    tmp=Conv2D(filters=300,padding="same",kernel_size=(1,7))(tmp)
    tmp=BatchNormalization(axis=-1)(tmp)
    added=Add()([x,tmp])
    out=Activation('relu')(added)
    return out

    
#note: functional model
def getModelGivenModelOptionsAndWeightInits(args):
    np.random.seed(args.seed)
    w0=args.w0
    w1=args.w1
    ntasks=args.ntasks
    
    K.set_image_data_format('channels_last')
    print(K.image_data_format())


    inputs=Input(shape=(1,1000,4))
    x=Conv2D(filters=300,kernel_size=(1,11))(inputs)
    x=BatchNormalization(axis=-1)(x)
    x=Activation('relu')(x)

    #add 2 x 1st resnet blocks
    x=l1_block(x)
    x=MaxPooling2D(pool_size=(1,4))(x)
    
    #Flatten
    x=Flatten()(x)

    #fully connected 1
    x=Dense(1000)(x)
    x=BatchNormalization(axis=-1)(x)
    x=Activation("relu")(x)
    x=Dropout(0.3)(x)

    #fully conected 2
    x=Dense(1000)(x)
    x=BatchNormalization(axis=-1)(x)
    x=Activation("relu")(x)
    x=Dropout(0.3)(x)
    
    #the output layer

    x=Dense(ntasks)(x)
    outputs=Activation("sigmoid")(x)
    
    model=Model(inputs=inputs,outputs=outputs)

    adam = keras.optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
    print("compiling!")
    import kerasAC.custom_losses
    if w0!=None:
        loss=kerasAC.custom_losses.get_weighted_binary_crossentropy(w0_weights=w0,w1_weights=w1)
    else:
        loss="binary_crossentropy"
    model.compile(optimizer=adam,loss=loss,metrics=[positive_accuracy,negative_accuracy,precision,recall])
    return model



import numpy as np
from kerasAC.metrics import *
import keras;
from keras.models import Model
from keras.layers import Input
from keras.layers.core import Dropout, Reshape, Dense, Activation, Flatten
from keras.layers.convolutional import Conv2D, MaxPooling2D
from keras.layers.merge import Add 
from keras.optimizers import Adadelta, SGD, RMSprop;
import keras.losses;
from keras.constraints import maxnorm;
from keras.layers.normalization import BatchNormalization
from keras.regularizers import l1, l2
from keras import backend as K

def l1_block(x):
    tmp=Conv2D(filters=64,kernel_size=(1,3),padding="same")(x)
    tmp=BatchNormalization(axis=-1)(tmp)
    tmp=Activation('relu')(tmp)
    tmp=Conv2D(filters=64,padding="same",kernel_size=(1,3))(tmp)
    tmp=BatchNormalization(axis=-1)(tmp)
    added=Add()([x,tmp])
    out=Activation('relu')(added)
    return out

def l2_block(x):
    tmp=Conv2D(filters=128,kernel_size=(1,7),padding="same")(x)
    tmp=BatchNormalization(axis=-1)(tmp)
    tmp=Activation('relu')(tmp)
    tmp=Conv2D(filters=128,kernel_size=(1,7),padding="same")(tmp)
    tmp=BatchNormalization(axis=-1)(tmp)
    added=Add()([x,tmp])
    out=Activation('relu')(added)
    return out

def l3_block(x):
    tmp=Conv2D(filters=200,kernel_size=(1,7),padding="same")(x)
    tmp=BatchNormalization(axis=-1)(tmp)
    tmp=Activation('relu')(tmp)
    tmp=Conv2D(filters=200,kernel_size=(1,3),padding="same")(tmp)
    tmp=BatchNormalization(axis=-1)(tmp)
    tmp=Activation('relu')(tmp)
    tmp=Conv2D(filters=200,kernel_size=(1,3),padding="same")(tmp)
    tmp=BatchNormalization(axis=-1)(tmp)
    added=Add()([x,tmp])
    out=Activation('relu')(added)
    return out


def l4_block(x):
    tmp=Conv2D(filters=200,kernel_size=(1,7),padding="same")(x)
    tmp=BatchNormalization(axis=-1)(tmp)
    tmp=Activation('relu')(tmp)
    tmp=Conv2D(filters=200,kernel_size=(1,7),padding="same")(tmp)
    tmp=BatchNormalization(axis=-1)(tmp)
    added=Add()([x,tmp])
    out=Activation('relu')(added)
    return out


#note: functional model
def getModelGivenModelOptionsAndWeightInits(args):
    #read in the args
    w0=args.w0
    w1=args.w1
    ntasks=args.ntasks
    seed=args.seed
    
    np.random.seed(seed)
    K.set_image_data_format('channels_last')
    print(K.image_data_format())

    #start with 2 conv layers
    inputs=Input(shape=(1,1000,4))
    x=Conv2D(filters=48,kernel_size=(1,3))(inputs)
    x=BatchNormalization(axis=-1)(x)
    x=Activation('relu')(x)

    x=Conv2D(filters=64,kernel_size=(1,3))(x)
    x=BatchNormalization(axis=-1)(x)
    x=Activation('relu')(x)

    #add 2 x 1st resnet blocks
    x=l1_block(x)
    x=l1_block(x)

    #add a conv layer
    x=Conv2D(filters=128,kernel_size=(1,3))(x)
    x=BatchNormalization(axis=-1)(x)
    x=Activation('relu')(x)

    #add 2 x 2nd resnet blocks
    x=l2_block(x)
    x=l2_block(x)
    
    #add maxpool
    x=MaxPooling2D(pool_size=(1,3))(x)

    #add a conv layer 
    x=Conv2D(filters=200,kernel_size=(1,1))(x)
    x=BatchNormalization(axis=-1)(x)
    x=Activation('relu')(x)


    #add 2 x 3rd resnet block
    x=l3_block(x)
    x=l3_block(x)

    #add maxpool
    x=MaxPooling2D(pool_size=(1,4))(x)

    
    #add 2 x 4th resnet block
    x=l4_block(x)
    x=l4_block(x)

    
    #add a conv layer 
    x=Conv2D(filters=200,kernel_size=(1,7))(x)
    x=BatchNormalization(axis=-1)(x)
    x=Activation('relu')(x)
    
    #add maxpool
    x=MaxPooling2D(pool_size=(1,4))(x)

    #Flatten
    x=Flatten()(x)

    #fully connected 1
    x=Dense(1000)(x)
    x=BatchNormalization(axis=-1)(x)
    x=Activation("relu")(x)
    x=Dropout(0.3)(x)

    #fully conected 2
    x=Dense(1000)(x)
    x=BatchNormalization(axis=-1)(x)
    x=Activation("relu")(x)
    x=Dropout(0.3)(x)
    
    #the output layer

    x=Dense(ntasks)(x)
    outputs=Activation("sigmoid")(x)
    
    model=Model(inputs=inputs,outputs=outputs)

    adam = keras.optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
    print("compiling!")
    if w0!=None:
        import kerasAC.custom_losses
        loss=kerasAC.custom_losses.get_weighted_binary_crossentropy(w0_weights=w0,w1_weights=w1)
    else:
        loss="binary_crossentropy"
    model.compile(optimizer=adam,loss=loss,metrics=[positive_accuracy,negative_accuracy,precision,recall])
    return model



import tensorflow.keras as keras

class LossHistory(keras.callbacks.Callback):
    
    def __init__(self,model_output_path_logs_name,to_track):
        self.model_output_path_logs_name=model_output_path_logs_name
        self.to_track=to_track
        self.outf=open(self.model_output_path_logs_name,'w')
        self.outf.write('Epoch\tBatch\t'+'\t'.join(self.to_track)+'\n')
        keras.callbacks.Callback.__init__(self)

    def on_train_begin(self, logs={}):
        self.losses ={}

    def on_epoch_begin(self,epoch, logs={}):
        self.losses[epoch]={}
        for trackable in self.to_track:
            self.losses[epoch][trackable]=[]         
        self.cur_epoch=epoch
        
    def on_batch_end(self, batch, logs={}):
        for trackable in self.to_track:
            self.losses[self.cur_epoch][trackable].append(logs.get(trackable))
            
    def on_epoch_end(self,epoch,logs={}):
        marker=self.to_track[0] 
        num_batches=len(self.losses[self.cur_epoch][marker])
        for i in range(num_batches):
            self.outf.write(str(epoch)+'\t'+str(i))
            for trackable in self.to_track:
                self.outf.write('\t'+str(self.losses[epoch][trackable][i]))
            self.outf.write('\n')
        
    def on_train_end(self,logs={}):
        self.outf.close()



#generates hg19_splits for cross-validation
#hg19 (i.e. /mnt/data/annotations/by_release/hg19.GRCh37/hg19.chrom.sizes). Any chromosome from this chrom.sizes file that is not in the test/validation split is assumed to be in the training split (only considering chroms 1 - 22, X, Y
hg19_chroms=['chr1','chr2','chr3','chr4','chr5','chr6','chr7','chr8','chr9','chr10','chr11','chr12','chr13','chr14','chr15','chr16','chr17','chr18','chr19','chr20','chr21','chr22','chrX','chrY']
hg19_splits=dict()

hg19_splits[0]={'test':['chr1'],
                'valid':['chr10','chr8'],
                'train':['chr2','chr3','chr4','chr5','chr6','chr7','chr9','chr11','chr12','chr13','chr14','chr15','chr16','chr17','chr18','chr19','chr20','chr21','chr22','chrX','chrY']}

hg19_splits[1]={'test':['chr19','chr2'],
                'valid':['chr1'],
                'train':['chr3','chr4','chr5','chr6','chr7','chr8','chr9','chr10','chr11','chr12','chr13','chr14','chr15','chr16','chr17','chr18','chr20','chr21','chr22','chrX','chrY']}

hg19_splits[2]={'test':['chr3','chr20'],
                'valid':['chr19','chr2'],
                'train':['chr1','chr4','chr5','chr6','chr7','chr8','chr9','chr10','chr11','chr12','chr13','chr14','chr15','chr16','chr17','chr18','chr21','chr22','chrX','chrY']}

hg19_splits[3]={'test':['chr13','chr6','chr22'],
                'valid':['chr3','chr20'],
                'train':['chr1','chr2','chr4','chr5','chr7','chr8','chr9','chr10','chr11','chr12','chr14','chr15','chr16','chr17','chr18','chr19','chr21','chrX','chrY']}

hg19_splits[4]={'test':['chr5','chr16','chrY'],
                'valid':['chr13','chr6','chr22'],
                'train':['chr1','chr2','chr3','chr4','chr7','chr8','chr9','chr10','chr11','chr12','chr14','chr15','chr17','chr18','chr19','chr20','chr21','chrX']}

hg19_splits[5]={'test':['chr4','chr15','chr21'],
                'valid':['chr5','chr16','chrY'],
                'train':['chr1','chr2','chr3','chr6','chr7','chr8','chr9','chr10','chr11','chr12','chr13','chr14','chr17','chr18','chr19','chr20','chr22','chrX']}

hg19_splits[6]={'test':['chr7','chr18','chr14'],
                'valid':['chr4','chr15','chr21'],
                'train':['chr1','chr2','chr3','chr5','chr6','chr8','chr9','chr10','chr11','chr12','chr13','chr16','chr17','chr19','chr20','chr22','chrX','chrY']}

hg19_splits[7]={'test':['chr11','chr17','chrX'],
                'valid':['chr7','chr18','chr14'],
                'train':['chr1','chr2','chr3','chr4','chr5','chr6','chr8','chr9','chr10','chr12','chr13','chr15','chr16','chr19','chr20','chr21','chr22','chrY']}

hg19_splits[8]={'test':['chr12','chr9'],
                'valid':['chr11','chr17','chrX'],
                'train':['chr1','chr2','chr3','chr4','chr5','chr6','chr7','chr8','chr10','chr13','chr14','chr15','chr16','chr18','chr19','chr20','chr21','chr22','chrY']}

hg19_splits[9]={'test':['chr10','chr8'],
                'valid':['chr12','chr9'],
                'train':['chr1','chr2','chr3','chr4','chr5','chr6','chr7','chr11','chr13','chr14','chr15','chr16','chr17','chr18','chr19','chr20','chr21','chr22','chrX','chrY']}


#Note: the splits for hg19 and hg38 are the same, as are the chromosomes used for training models. 
splits=dict()
splits['hg19']=hg19_splits 
splits['hg38']=hg19_splits
chroms=dict()
chroms['hg19']=hg19_chroms
chroms['hg38']=hg19_chroms


mm10_chroms = ['chr1','chr2','chr3','chr4','chr5','chr6','chr7','chr8','chr9','chr10','chr11','chr12','chr13','chr14','chr15','chr16','chr17','chr18','chr19','chrX','chrY']
mm10_splits = dict()

mm10_splits[0] = {'test':['chr1'],
                  'valid':['chr10','chr8'],
                  'train':['chr2','chr3','chr4','chr5','chr6','chr7','chr9','chr11','chr12','chr13','chr14','chr15','chr16','chr17','chr18','chr19','chrX','chrY']}

mm10_splits[1] = {'test':['chr19','chr2'],
                  'valid':['chr1'],
                  'train':['chr3','chr4','chr5','chr6','chr7','chr8','chr9','chr10','chr11','chr12','chr13','chr14','chr15','chr16','chr17','chr18','chrX','chrY']}

mm10_splits[2] = {'test':['chr3'],
                  'valid':['chr19','chr2'],
                  'train':['chr1','chr4','chr5','chr6','chr7','chr8','chr9','chr10','chr11','chr12','chr13','chr14','chr15','chr16','chr17','chr18','chrX','chrY']}

mm10_splits[3] = {'test':['chr13','chr6'],
                  'valid':['chr3'],
                  'train':['chr1','chr2','chr4','chr5','chr6','chr7','chr8','chr9','chr10','chr11','chr12','chr14','chr15','chr17','chr18','chr19','chrX','chrY']}

mm10_splits[4] = {'test':['chr5','chr16','chrY'],
                  'valid':['chr13','chr6'],
                  'train':['chr1','chr2','chr3','chr4','chr7','chr8','chr9','chr10','chr11','chr12','chr14','chr15','chr17','chr18','chr19','chrX']}

mm10_splits[5] = {'test':['chr4','chr15'],
                  'valid':['chr5','chr16','chrY'],
                  'train':['chr1','chr2','chr3','chr6','chr7','chr8','chr9','chr10','chr11','chr12','chr13','chr14','chr17','chr18','chr19','chrX']}

mm10_splits[6] = {'test':['chr7','chr18','chr14'],
                  'valid':['chr4','chr15'],
                  'train':['chr1','chr2','chr3','chr5','chr6','chr8','chr9','chr10','chr11','chr12','chr13','chr16','chr17','chr19','chrX','chrY']}

mm10_splits[7] = {'test':['chr11','chr17','chrX'],
                  'valid':['chr7','chr18','chr14'],
                  'train':['chr1','chr2','chr3','chr4','chr5','chr6','chr8','chr9','chr10','chr12','chr13','chr15','chr16','chr19','chrY']}

mm10_splits[8] = {'test':['chr12','chr9'],
                  'valid':['chr11','chr17','chrX'],
                  'train':['chr1','chr2','chr3','chr4','chr5','chr6','chr7','chr8','chr10','chr13','chr14','chr15','chr16','chr18','chr19','chrY']}

mm10_splits[9] = {'test':['chr10','chr8'],
                  'valid':['chr12','chr9'],
                  'train':['chr1','chr2','chr3','chr4','chr5','chr6','chr7','chr11','chr13','chr14','chr15','chr16','chr17','chr18','chr19','chrX','chrY']}

#mm9 & mm10 splits are the same
splits['mm10']=mm10_splits
splits['mm9']=mm10_splits
chroms['mm10']=mm10_splits
chroms['mm9']=mm10_splits


def get_chroms(args,split):
    if split=="train":
        if args.train_chroms is not None:
            return args.train_chroms
        else:
            assert args.genome is not None
            assert args.fold is not None
            return splits[args.genome][args.fold]['train']
    elif split=='valid':
        if args.validation_chroms is not None:
            return args.validation_chroms
        else:
            assert args.genome is not None
            assert args.fold is not None
            return splits[args.genome][args.fold]['valid']
    elif split=="test":
        if args.predict_chroms is not None:
            return args.predict_chroms
        else:
            assert args.genome is not None
            assert args.fold is not None
            return splits[args.genome][args.fold]['test']
    else:
        raise Exception("invalid split specified, must be one of train,valid,test; you gave:"+str(split))
    

def get_bed_regions_for_fold_split(bed_regions,genome,fold,split):
    chroms_to_keep=splits[genome][int(fold)][split]
    bed_regions_to_keep=bed_regions[bed_regions[0].isin(chroms_to_keep)]
    print("got split:"+str(split)+" for fold:"+str(fold) +" for bed regions:"+str(bed_regions_to_keep.shape))
    return bed_regions_to_keep

    
    



import tensorflow as tf
from .splits import *
from .config import args_object_from_args_dict
from .train import *
from .predict_hdf5 import *
from .interpret import *
from .performance_metrics.performance_metrics import * 
import argparse
import pdb

def parse_args():
    parser=argparse.ArgumentParser(description='Provide model files  & a dataset, get model predictions')
    parser.add_argument("--model_prefix",help="output model file that is generated at the end of training (in hdf5 format)")
    parser.add_argument("--assembly")
    parser.add_argument("--splits",nargs="+",default=None,type=int)
    parser.add_argument("--seed",type=int,default=1234)
    parser.add_argument("--use_multiprocessing",action='store_true',default=False)
    parser.add_argument("--multi_gpu",action='store_true',default=False)
    input_data_path=parser.add_argument_group("input_data_path")
    input_data_path.add_argument("--index_data_path",default=None,help="seqdataloader output hdf5, or tsv file containing binned labels")
    input_data_path.add_argument("--index_train_path",default=None,help="seqdataloader output hdf5, or tsv file containing binned labels for the training split")
    input_data_path.add_argument("--index_valid_path",default=None,help="seqdataloader output hdf5, or tsv file containing binned labels for the validation split")
    input_data_path.add_argument("--index_tasks",nargs="*",default=None)
    input_data_path.add_argument("--input_data_path",nargs="+",default=None,help="seq or path to seqdataloader hdf5")
    input_data_path.add_argument("--input_train_path",nargs="+",default=None,help="seq or seqdataloader hdf5")
    input_data_path.add_argument("--input_valid_path",nargs="+",default=None,help="seq or seqdataloader hdf5")
    input_data_path.add_argument("--output_data_path",nargs="+",default=None,help="path to seqdataloader hdf5")
    input_data_path.add_argument("--output_train_path",nargs="+",default=None,help="seqdataloader hdf5")
    input_data_path.add_argument("--output_valid_path",nargs="+",default=None,help="seqdataloader hdf5")    
    input_data_path.add_argument('--variant_bed',default=None)
    input_data_path.add_argument('--ref_fasta')

    tiledbgroup=parser.add_argument_group("tiledb")
    tiledbgroup.add_argument("--tdb_outputs",nargs="+")
    tiledbgroup.add_argument("--tdb_output_source_attribute",nargs="+",default="fc_bigwig",help="tiledb attribute for use in label generation i.e. fc_bigwig")
    tiledbgroup.add_argument("--tdb_output_flank",nargs="+",type=int,help="flank around bin center to use in generating outputs")
    tiledbgroup.add_argument("--tdb_output_aggregation",nargs="+",default=None,help="method for output aggreagtion; one of None, 'avg','max'")
    tiledbgroup.add_argument("--tdb_output_transformation",nargs="+",default=None,help="method for output transformation; one of None, 'log','log10','asinh'")    
    tiledbgroup.add_argument("--tdb_inputs",nargs="+")
    tiledbgroup.add_argument("--tdb_input_source_attribute",nargs="+",help="attribute to use for generating model input, or 'seq' for one-hot-encoded sequence")
    tiledbgroup.add_argument("--tdb_input_flank",nargs="+",type=int,help="length of sequence around bin center to use for input")
    tiledbgroup.add_argument("--tdb_input_aggregation",nargs="+",default=None,help="method for input aggregation; one of 'None','avg','max'")
    tiledbgroup.add_argument("--tdb_input_transformation",nargs="+",default=None,help="method for input transformation; one of None, 'log','log10','asinh'")
    tiledbgroup.add_argument("--tdb_indexer",default=None,help="tiledb paths for each input task")
    tiledbgroup.add_argument("--tdb_partition_attribute_for_upsample",default="idr_peak",help="tiledb attribute to use for upsampling, i.e. idr_peak")
    tiledbgroup.add_argument("--tdb_partition_thresh_for_upsample",type=float,default=1,help="values >= partition_thresh_for_upsample within the partition_attribute_for_upsample will be upsampled during training")
    tiledbgroup.add_argument("--upsample_ratio_list_predict",type=float,nargs="*")    
    tiledbgroup.add_argument("--chrom_sizes",default=None,help="chromsizes file for use with tiledb generator")
    tiledbgroup.add_argument("--tiledb_stride",type=int,default=1)

    input_filtering_params=parser.add_argument_group("input_filtering_params")    
    input_filtering_params.add_argument('--center_on_summit',default=False,action='store_true',help="if this is set to true, the peak will be centered at the summit (must be last entry in bed file or hammock) and expanded args.flank to the left and right")
    
    output_params=parser.add_argument_group("output_params")
    output_params.add_argument('--predictions_and_labels_hdf5',help='name of hdf5 to save predictions',default=None)
    output_params.add_argument('--performance_metrics_classification_file',nargs="+", help='file name to save accuracy metrics; accuracy metrics not computed if file not provided',default=None)
    output_params.add_argument('--performance_metrics_regression_file',nargs="+", help='file name to save accuracy metrics; accuracy metrics not computed if file not provided',default=None)
    output_params.add_argument('--performance_metrics_profile_file',nargs="+", help='file name to save accuracy metrics; accuracy metrics not computed if file not provided',default=None)
    
    calibration_params=parser.add_argument_group("calibration_params")
    calibration_params.add_argument("--calibrate_classification",action="store_true",default=False)
    calibration_params.add_argument("--calibrate_regression",action="store_true",default=False)        
    
    model_params=parser.add_argument_group("model_params")
    model_params.add_argument('--weights',help='weights file for the model')
    model_params.add_argument('--yaml',help='yaml file for the model')
    model_params.add_argument('--json',help='json file for the model')
    model_params.add_argument('--functional',default=False,help='use this flag if your model is a functional model',action="store_true")
    model_params.add_argument('--squeeze_input_for_gru',action='store_true')
    model_params.add_argument("--expand_dims",default=False,action="store_true")
    model_params.add_argument("--num_inputs",type=int)
    model_params.add_argument("--num_outputs",type=int)

    arch_params=parser.add_argument_group("arch_params")
    arch_params.add_argument("--from_checkpoint_arch",default=None)
    arch_params.add_argument("--architecture_spec",type=str,default="basset_architecture_multitask")
    arch_params.add_argument("--architecture_from_file",type=str,default=None)
    arch_params.add_argument("--num_tasks",type=int)
    arch_params.add_argument("--tasks",nargs="*",default=None)
    arch_params.add_argument("--task_indices",nargs="*",default=None,help="list of tasks to train on, by index of their position in tdb matrix")
        
    parallelization_params=parser.add_argument_group("parallelization")
    parallelization_params.add_argument("--threads",type=int,default=1)
    parallelization_params.add_argument("--max_queue_size",type=int,default=100)
    parallelization_params.add_argument("--num_gpus",type=int,default=1)

    snp_params=parser.add_argument_group("snp_params")
    snp_params.add_argument("--vcf_file",default=None)
    snp_params.add_argument("--global_vcf",action="store_true")
    snp_params.add_argument('--background_freqs',default=None)
    snp_params.add_argument('--flank',default=500,type=int)
    snp_params.add_argument('--mask',default=10,type=int)
    snp_params.add_argument('--ref_col',type=int,default=None)
    snp_params.add_argument('--alt_col',type=int,default=None)

    train_val_splits=parser.add_argument_group("train_val_splits")
    train_val_splits.add_argument("--num_train",type=int,default=700000)
    train_val_splits.add_argument("--num_valid",type=int,default=150000)

    batch_params=parser.add_argument_group("batch_params")
    batch_params.add_argument("--batch_size",type=int,default=1000)
    batch_params.add_argument("--revcomp",action="store_true")
    batch_params.add_argument("--label_transformer",nargs="+",default=None,help="transformation to apply to label values")
    batch_params.add_argument("--upsample_thresh_list_train",type=float,nargs="*",default=None)
    batch_params.add_argument("--upsample_ratio_list_train",type=float,nargs="*",default=None)
    batch_params.add_argument("--upsample_thresh_list_eval",type=float,nargs="*",default=None)
    batch_params.add_argument("--upsample_ratio_list_eval",type=float,nargs="*",default=None)

    
    weights_params=parser.add_argument_group("weights_params")
    weights_params.add_argument("--init_weights",default=None)
    weights_params.add_argument('--w1',nargs="*", type=float, default=None)
    weights_params.add_argument('--w0',nargs="*", type=float, default=None)
    weights_params.add_argument("--w1_w0_file",default=None)
    weights_params.add_argument("--save_w1_w0", default=None,help="output text file to save w1 and w0 to")
    weights_params.add_argument("--weighted",action="store_true",help="separate task-specific weights denoted with w1, w0 args are to be used")
    weights_params.add_argument("--from_checkpoint_weights",default=None)

    epoch_params=parser.add_argument_group("epoch_params")
    epoch_params.add_argument("--epochs",type=int,default=40)
    epoch_params.add_argument("--patience",type=int,default=3)
    epoch_params.add_argument("--patience_lr",type=int,default=2,help="number of epochs with no drop in validation loss after which to reduce lr")
    epoch_params.add_argument("--shuffle_epoch_start",action='store_true',default=False)
    epoch_params.add_argument("--shuffle_epoch_end",action='store_true',default=False)

    vis_params=parser.add_argument_group("visualization")            
    vis_params.add_argument("--tensorboard",action="store_true")
    vis_params.add_argument("--tensorboard_logdir",default="logs")
    vis_params.add_argument("--trackables",nargs="*",default=['loss','val_loss'], help="list of things to track per batch, such as logcount_predictions_loss,loss,profile_predictions_loss,val_logcount_predictions_loss,val_loss,val_profile_predictions_loss")
    return parser.parse_args()

def cross_validate(args):
    if type(args)==type({}):
        args=args_object_from_args_dict(args) 

    #run training on each of the splits
    if args.assembly not in splits:
        raise Exception("You did not provide the args.assembly flag, or you provided an unsupported genome assembly:"+str(args.assembly)+". Supported assemblies include:"+str(splits.keys())+"; add splits for this assembly to splits.py file")
    args_dict=vars(args)
    print(args_dict) 
    base_model_file=args_dict['model_prefix']
    base_performance_classification_file=args_dict['performance_metrics_classification_file']
    base_performance_regression_file=args_dict['performance_metrics_regression_file']
    base_performance_profile_file=args_dict['performance_metrics_profile_file']

    base_predictions_and_labels_hdf5=args_dict['predictions_and_labels_hdf5']
    base_init_weights=args_dict['init_weights'] 

    all_splits=splits[args.assembly]
    if args.splits!=None:
        all_splits=args.splits 
    for split in all_splits: 
        print("Starting split:"+str(split))

        test_chroms=splits[args.assembly][split]['test']
        validation_chroms=splits[args.assembly][split]['valid']
        train_chroms=list(set(chroms[args.assembly])-set(test_chroms+validation_chroms))

        #convert args to dict
        args_dict=vars(args)
        args_dict['train_chroms']=train_chroms
        args_dict['validation_chroms']=validation_chroms
        
        if base_init_weights is not None:
            args_dict['init_wights']=base_init_weights+"."+str(split)
        #set the training arguments specific to this fold 
        args_dict['model_prefix']=base_model_file+"."+str(split)
        args_dict['tdb_array']=None
        args_dict['load_model_hdf5']=None
        print("Training model on split"+str(split)) 
        train(args_dict)
        
        #set the prediction arguments specific to this fold
        if args.save_w1_w0!=None:
            args_dict["w1_w0_file"]=args.save_w1_w0
        if base_predictions_and_labels_hdf5!=None:
            args_dict['predictions_and_labels_hdf5']=base_predictions_and_labels_hdf5+"."+str(split)
            print(args_dict['predictions_and_labels_hdf5'])
        args_dict['predict_chroms']=test_chroms
        print("Calculating predictions on the test fold in split:"+str(split)) 
        predict(args_dict)

        #score the predictions
        print("scoring split:"+str(split))
        preds_to_score=[]
        labels_to_score=[]
        for cur_output in range(args.num_outputs):
            cur_preds=base_predictions_and_labels_hdf5+"."+str(split)+".predictions."+str(cur_output)
            cur_labels=base_predictions_and_labels_hdf5+"."+str(split)+".labels."+str(cur_output)
            preds_to_score.append(cur_preds)
            labels_to_score.append(cur_labels)
                
        if base_performance_classification_file is not None:
            args_dict['performance_metrics_classification_file']=[perf_file+"."+str(split) for perf_file in base_performance_classification_file]
        elif base_performance_regression_file is not None:
            args_dict['performance_metrics_regression_file']=[perf_file +'.'+str(split) for perf_file in base_performance_regression_file]
        elif base_performance_profile_file is not None:
            args_dict['performance_metrics_profile_file']=[perf_file +'.'+str(split) for perf_file in base_performance_profile_file]
        print(preds_to_score)
        print(labels_to_score)
        args_dict['predictions_hdf5']=preds_to_score
        args_dict['labels_hdf5']=labels_to_score
        get_performance_metrics(args)
        
        
def main():
    args=parse_args()
    cross_validate(args) 

if __name__=="__main__":
    main()



import pdb 
from kerasAC.custom_losses import * 
from kerasAC.metrics import recall, specificity, fpr, fnr, precision, f1
from keras.models import load_model
import sys 
custom_objects={"recall":recall,
                "sensitivity":recall,
                "specificity":specificity,
                "fpr":fpr,
                "fnr":fnr,
                "precision":precision,
                "f1":f1,
                "ambig_binary_crossentropy":ambig_binary_crossentropy,
                "ambig_mean_absolute_error":ambig_mean_absolute_error,
                "ambig_mean_squared_error":ambig_mean_squared_error}
model_path=sys.argv[1]        
model=load_model(model_path,custom_objects=custom_objects)
print(model.summary())
pdb.set_trace()



import tensorflow as tf
import tensorflow_probability as tfp

import pdb 
import numpy as np 
def get_weighted_binary_crossentropy(w0_weights, w1_weights,ambig_val=np.nan):
    # Compute the task-weighted cross-entropy loss, where every task is weighted by 1 - (fraction of non-ambiguous examples that are positive)
    # In addition, weight everything with label -1 to 0
    import numpy as np 
    w0_weights=np.array(w0_weights);
    w1_weights=np.array(w1_weights);
    def weighted_binary_crossentropy(y_true,y_pred):
        weightsPerTaskRep = y_true*w1_weights[None,:] + (1-y_true)*w0_weights[None,:]
        nonAmbig=tf.math.logical_not(tf.math.is_nan(y_true))
        nonAmbigTimesWeightsPerTask = tf.boolean_mask(weightsPerTaskRep,nonAmbig)
        return tf.mean(tf.binary_crossentropy(tf.boolean_mask(y_true,nonAmbig),tf.boolean_mask(y_pred,nonAmbig))*nonAmbigTimesWeightsPerTask, axis=-1);
    return weighted_binary_crossentropy; 

def ambig_binary_crossentropy(y_true,y_pred):
    nonAmbig=tf.math.logical_not(tf.math.is_nan(y_true))
    return tf.mean(tf.binary_crossentropy(tf.boolean_mask(y_true,nonAmbig), tf.boolean_mask(y_pred,nonAmbig)), axis=-1);

def ambig_mean_squared_error(y_true, y_pred):
    nonAmbig=tf.math.logical_not(tf.math.is_nan(y_true))
    return tf.mean(tf.square(tf.boolean_mask(y_pred,nonAmbig) - tf.boolean_mask(y_true,nonAmbig)), axis=-1)

def ambig_mean_absolute_error(y_true, y_pred):
    nonAmbig=tf.math.logical_not(tf.math.is_nan(y_true))
    return tf.mean(tf.abs(tf.boolean_mask(y_pred,nonAmbig) - tf.boolean_mask(y_true,nonAmbig)), axis=-1)

def ambig_log_poisson(y_true,y_pred):
    nonAmbig=tf.math.logical_not(tf.math.is_nan(y_true))
    return tf.nn.log_poisson_loss(nonAmbig,tf.log(y_pred),compute_full_loss=True)


#PROFILE MODEL LOSSES #
def get_loss_weights(tdb_path,chrom,label_attribute,ambig_attribute,upsample_attribute,tdb_partition_thresh_for_upsample):
    import tiledb
    from kerasAC.tiledb_config import get_default_config
    import pdb 
    tdb_config=get_default_config()
    ctx=tiledb.Ctx(tdb_config)
    tdb_array=tiledb.DenseArray(tdb_path+"."+chrom,mode='r',ctx=ctx)
    print("opened:"+tdb_path+"."+chrom+" for reading")
    vals=tdb_array[:]
    print("got tdb vals")
    #label_attribute

def multinomial_nll(true_counts, logits):
    """Compute the multinomial negative log-likelihood
    Args:
      true_counts: observed count values
      logits: predicted logit values
    """
    counts_per_example = tf.reduce_sum(true_counts, axis=-1)
    dist = tfp.distributions.Multinomial(total_count=counts_per_example,
                                         logits=logits)
    return (-tf.reduce_sum(dist.log_prob(true_counts)) / 
            tf.cast(tf.shape(true_counts)[0], dtype=tf.float32))

def custom_mse(y_true, y_pred):
    # calculating squared difference between target and predicted values 
    loss = tf.square(y_pred - y_true)  # (batch_size, 2)
    # summing both loss values along batch dimension 
    loss = tf.sum(loss, axis=1)        # (batch_size,)
    return loss

#from https://github.com/kundajelab/basepair/blob/cda0875571066343cdf90aed031f7c51714d991a/basepair/losses.py#L87
class MultichannelMultinomialNLL(object):
    def __init__(self, n, weights=None):
        self.__name__ = "MultichannelMultinomialNLL"
        self.n = n
        if weights is None:
            self.weights = [1]*self.n
        else:
            self.weights = weights
        #print(self.weights, self.n)


    def __call__(self, true_counts, logits):
        for i in range(self.n):
            loss = multinomial_nll(true_counts[..., i], logits[..., i])
            #print(loss)
            if i == 0:
                total = self.weights[i]*loss
            else:
                total += self.weights[i]*loss
        return total

    def get_config(self):
        return {"n": self.n, "weights":self.weights}

class MultichannelMultinomialMSE(object):
    def __init__(self, n, weights=None):
        self.__name__ = "MultichannelMultinomialMSE"
        self.n = n
        if weights is None:
            self.weights = [1]*self.n
        else:
            self.weights = weights
        #print(self.weights, self.n)
        self.mse = tf.keras.losses.MeanSquaredError()

    def __call__(self, true_counts, logits):
        for i in range(self.n):
            #print(true_counts[..., i], logits[..., i])
            loss = self.mse(true_counts[..., i], logits[..., i])
            #print(loss, self.n)
            if i == 0:
                total = self.weights[i]*loss
            else:
                total += self.weights[i]*loss
        return total

    def get_config(self):
        return {"n": self.n, "weights":self.weights}




#utilities for syncing to/from s3 file
import boto3
import tempfile
import os 
global to_clean
#delete tmp files at the end of the run
to_clean=[]

def s3_string_parse(s3_string):
    #sync log, model hdf5, weight file, arch file
    s3_client=boto3.client(service_name='s3')
    bucket=s3_string.lstrip('s3://').split('/')[0]
    filename="/".join(s3_string.lstrip("s3://").split("/")[1::])
    return bucket,filename

#target file for local download is either provided or inferred as the basenme
# of the s3 file path 
def download_s3_file(s3_source,s3_target=None):
    s3 = boto3.client('s3')
    bucket,s3_file=s3_string_parse(s3_source)
    if s3_target is None:
        s3_target=s3_file.split('/')[-1]
    s3.download_file(bucket, s3_file, s3_target)
    print("download of "+s3_source+ " to " + s3_target + " is complete")
    to_clean.append(s3_target)
    return s3_target

def upload_s3_file(s3_target,source_filename):
    s3 = boto3.client('s3')
    bucket,s3_file=s3_string_parse(s3_target)
    s3.upload_file(source_filename,bucket,s3_file)
    to_clean.append(source_filename)
    print("upload of "+s3_target+" is complete")


def read_s3_file_contents(s3_string):
    s3=boto3.resource('s3')
    bucket_name,itemname=s3_string_parse(s3_string)
    obj=s3.Object(bucket_name,itemname)
    contents=obj.get()['Body'].read().decode('utf-8').strip()
    return contents
    

def run_cleanup():
    to_clean_set=list(set(to_clean))
    print("to_clean_set:"+str(to_clean_set))
    if len(to_clean_set)>0:
        for f in to_clean_set:
            try:
                os.remove(f)
                print("deleted local file:"+str(f))
            except:
                continue
    return 
        



from __future__ import division, print_function, absolute_import
import tensorflow as tf
import importlib
import imp
import os
import tempfile
import argparse
import numpy as np
import h5py
from .s3_sync import *
from .generators.basic_generator import *
from .generators.tiledb_generator import *
from .custom_callbacks import * 
from .tiledb_config import *
from .get_model import *
from .splits import * 
from . import config
import pdb
from tensorflow.keras.callbacks import *
from tensorflow.keras.utils import multi_gpu_model
import gc
import multiprocessing
import pandas as pd
#multiprocessing.set_start_method('forkserver', force=True)
def parse_args():
    parser=argparse.ArgumentParser()
    
    parser.add_argument("--model_prefix",help="output model file that is generated at the end of training (in hdf5 format)")
    parser.add_argument("--seed",type=int,default=1234)    
    parser.add_argument("--num_inputs",type=int)
    parser.add_argument("--num_outputs",type=int)
    parser.add_argument("--use_multiprocessing",action='store_true',default=False)
    
    tiledbgroup=parser.add_argument_group('tiledb')
    tiledbgroup.add_argument("--tdb_array",help="name of tdb array to use")

    tiledbgroup.add_argument("--tdb_output_datasets",nargs="+",default=None,help="dataset column from db_ingest; comma separated across channels; space separated across outputs")
    tiledbgroup.add_argument("--tdb_output_source_attribute",nargs="+",help="tiledb attribute for use in label generation i.e. fc_bigwig")
    tiledbgroup.add_argument("--tdb_output_min",nargs="*", default=None)
    tiledbgroup.add_argument("--tdb_output_max",nargs="*", default=None)        
    tiledbgroup.add_argument("--tdb_output_flank",nargs="+",help="flank around bin center to use in generating outputs")
    tiledbgroup.add_argument("--tdb_output_aggregation",nargs="+",help="method for output aggregation; one of None, 'avg','max'")
    tiledbgroup.add_argument("--tdb_output_transformation",nargs="+",help="method for output transformation; one of None, 'log','log10','asinh'")
    tiledbgroup.add_argument("--tdb_transformation_pseudocount",type=float,default=0.001)

    tiledbgroup.add_argument("--tdb_input_datasets",nargs="+",default=None,help="dataset column from db_ingest; comma separated across channels; space separated across inputs")
    tiledbgroup.add_argument("--tdb_input_source_attribute",nargs="+",help="attribute to use for generating model input, or 'seq' for one-hot-encoded sequence")
    tiledbgroup.add_argument("--tdb_input_min",nargs="*", default=None)
    tiledbgroup.add_argument("--tdb_input_max",nargs="*", default=None)    
    tiledbgroup.add_argument("--tdb_input_flank",nargs="+",help="length of sequence around bin center to use for input")
    tiledbgroup.add_argument("--tdb_input_aggregation",nargs="+",help="method for input aggregation; one of 'None','avg','max'")
    tiledbgroup.add_argument("--tdb_input_transformation",nargs="+",help="method for input transformation; one of None, 'log','log10','asinh'")

    tiledbgroup.add_argument("--tdb_partition_attribute_for_upsample",default=None,help="tiledb attribute to use for upsampling, i.e. idr_peak")
    tiledbgroup.add_argument("--tdb_partition_thresh_for_upsample",type=float,default=None,help="values >= partition_thresh_for_upsample within the partition_attribute_for_upsample will be upsampled during training")
    tiledbgroup.add_argument("--tdb_partition_datasets_for_upsample",nargs="+")

    tiledbgroup.add_argument("--tdb_ambig_attribute",default=None,help="attribute indicating ambiguous regions to not train on")    

    
    input_data_path=parser.add_argument_group('input_data_path')
    input_data_path.add_argument("--index_data_path",default=None,help="seqdataloader output hdf5, or tsv file containing binned labels")
    input_data_path.add_argument("--index_train_path",default=None,help="seqdataloader output hdf5, or tsv file containing binned labels for the training split")
    input_data_path.add_argument("--index_valid_path",default=None,help="seqdataloader output hdf5, or tsv file containing binned labels for the validation split")
    input_data_path.add_argument("--index_tasks",nargs="*",default=None)
    input_data_path.add_argument("--hdf5_tasks",nargs="+") 
    input_data_path.add_argument("--input_data_path",nargs="+",default=None,help="seq or path to seqdataloader hdf5")
    input_data_path.add_argument("--input_train_path",nargs="+",default=None,help="seq or seqdataloader hdf5")
    input_data_path.add_argument("--input_valid_path",nargs="+",default=None,help="seq or seqdataloader hdf5")

    input_data_path.add_argument("--output_data_path",nargs="+",default=None,help="path to seqdataloader hdf5")
    input_data_path.add_argument("--output_train_path",nargs="+",default=None,help="seqdataloader hdf5")
    input_data_path.add_argument("--output_valid_path",nargs="+",default=None,help="seqdataloader hdf5")
    
    input_data_path.add_argument("--ref_fasta",default="/mnt/data/annotations/by_release/hg19.GRCh37/hg19.genome.fa")
    
    train_val_splits=parser.add_argument_group("train_val_splits")
    train_val_splits.add_argument("--train_chroms",nargs="*",default=None)
    train_val_splits.add_argument("--validation_chroms",nargs="*",default=None)
    train_val_splits.add_argument("--genome",default=None)
    train_val_splits.add_argument("--fold",type=int,default=None)
    train_val_splits.add_argument("--bed_regions",default=None,help="bed file for training/validation; train/splits are generated from the --fold and --genome arguments automatically")
    train_val_splits.add_argument("--bed_regions_train",default=None)
    train_val_splits.add_argument("--bed_regions_validate",default=None)
    train_val_splits.add_argument('--bed_regions_center',choices=['summit','center','random'],help="summit = center on peak summti in column 10 of narrowPeak file; center = center on the base halfway between start (col 2) and end (col 3); random = pick a random base between start and end for centering; this is useful for adding jitter when training")
    train_val_splits.add_argument("--bed_regions_jitter",type=int,default=1,help="number of positions within a bed region to use as interval centers; i.e. adds jitter for training/validation")
    train_val_splits.add_argument("--num_train",type=int,default=700000)
    train_val_splits.add_argument("--num_valid",type=int,default=150000)

    weights_params=parser.add_argument_group("weights_params")
    weights_params.add_argument("--load_model_hdf5")
    weights_params.add_argument("--weights",default=None)
    weights_params.add_argument("--init_weights",default=None,help="legacy, will be deprecated in next release") 
    weights_params.add_argument('--w1',nargs="*", type=float, default=None)
    weights_params.add_argument('--w0',nargs="*", type=float, default=None)
    weights_params.add_argument("--w1_w0_file",default=None)
    weights_params.add_argument("--save_w1_w0", default=None,help="output text file to save w1 and w0 to")
    weights_params.add_argument("--weighted",action="store_true",help="separate task-specific weights denoted with w1, w0 args are to be used")
    
    
    arch_params=parser.add_argument_group("arch_params")
    arch_params.add_argument("--json",default=None)
    arch_params.add_argument("--yaml",default=None)
    arch_params.add_argument("--architecture_spec",type=str,default="basset_architecture_multitask")
    arch_params.add_argument("--architecture_from_file",type=str,default=None)
    arch_params.add_argument("--model_params",type=str,default=None,help="2-column file with param name in column 1 and param value in column 2")
    arch_params.add_argument("--num_tasks",type=int)
    arch_params.add_argument("--datasets",nargs="*",default=None,help="list of tasks to train on, by name")
    arch_params.add_argument("--dataset_indices",nargs="*",default=None,help="list of tasks to train on, by index of their position in tdb matrix")
    
    batch_params=parser.add_argument_group("batch_params")
    batch_params.add_argument("--batch_size",type=int,default=1000)
    batch_params.add_argument("--revcomp",action="store_true")
    batch_params.add_argument("--squeeze_input_for_gru",action="store_true")
    batch_params.add_argument("--expand_dims",default=False,action="store_true",help="this is deprecated for tiledb, only supported for legacy compatibility with hdf5; ignored for tiledb")
    batch_params.add_argument("--upsample_thresh_list_train",type=float,nargs="*",default=None)
    batch_params.add_argument("--upsample_ratio_list_train",type=float,nargs="*",default=None)
    batch_params.add_argument("--upsample_thresh_list_eval",type=float,nargs="*",default=None)
    batch_params.add_argument("--upsample_ratio_list_eval",type=float,nargs="*",default=None)
    batch_params.add_argument("--upsample_threads",type=int,default=1)
    
    epoch_params=parser.add_argument_group("epoch_params")
    epoch_params.add_argument("--epochs",type=int,default=40)
    epoch_params.add_argument("--patience",type=int,default=3)
    epoch_params.add_argument("--patience_lr",type=int,default=2,help="number of epochs with no drop in validation loss after which to reduce lr")
    epoch_params.add_argument("--shuffle_epoch_end",action='store_true', default=True)
    
    #add functionality to train on individuals' allele frequencies
    snp_params=parser.add_argument_group("snp_params")
    snp_params.add_argument("--vcf_file",default=None)
    snp_params.add_argument("--global_vcf",action="store_true")

    parallelization_params=parser.add_argument_group("parallelization")
    parallelization_params.add_argument("--threads",type=int,default=1)
    parallelization_params.add_argument("--max_queue_size",type=int,default=100)
    parallelization_params.add_argument("--num_gpus",type=int,default=1)

    vis_params=parser.add_argument_group("visualization")            
    vis_params.add_argument("--tensorboard",action="store_true")
    vis_params.add_argument("--tensorboard_logdir",default="logs")
    vis_params.add_argument("--trackables",nargs="*",default=['loss','val_loss'], help="list of things to track per batch, such as logcount_predictions_loss,loss,profile_predictions_loss,val_logcount_predictions_loss,val_loss,val_profile_predictions_loss")
    return parser.parse_args()


def fit_and_evaluate(model,train_gen,valid_gen,args):
    #accomodate storage on s3
    if args.model_prefix.startswith('s3'):
        #store in local temporary file
        model_output_path_string=os.path.basename(args.model_prefix)
        model_output_path_hdf5=tempfile.NamedTemporaryFile(suffix=model_output_path_string+".hdf5")
        model_output_path_logs=tempfile.NamedTemporaryFile(suffix=model_output_path_string+".log")        
        model_output_path_arch=tempfile.NamedTemporaryFile(suffix=model_output_path_string+".arch")
        model_output_path_weights=tempfile.NamedTemporaryFile(suffix=model_output_path_string+".weights")
        model_output_path_hdf5_name=model_output_path_hdf5.name
        model_output_path_logs_name=model_output_path_logs.name
        model_output_path_arch_name=model_output_path_arch.name
        model_output_path_weights_name=model_output_path_weights.name
        
    else: 
        model_output_path_string = args.model_prefix
        model_output_path_hdf5_name=model_output_path_string+".hdf5"
        model_output_path_logs_name=model_output_path_string+".log"
        model_output_path_arch_name=model_output_path_string+".arch"
        model_output_path_weights_name=model_output_path_string+".weights"

    
    checkpointer = ModelCheckpoint(filepath=model_output_path_hdf5_name, verbose=1, save_best_only=True,save_weights_only=False)
    try:
        earlystopper = EarlyStopping(monitor='val_loss', patience=args.patience, verbose=1, restore_best_weights=True)
    except:
        #compatibility with keras <  2.2. 
        earlystopper = EarlyStopping(monitor='val_loss', patience=args.patience, verbose=1)
    history=LossHistory(model_output_path_logs_name+".batch",args.trackables)
    csvlogger = CSVLogger(model_output_path_logs_name, append = False)
    reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=0.4,patience=args.patience_lr, min_lr=0.00000001)
    cur_callbacks=[checkpointer,earlystopper,csvlogger,reduce_lr,history]
    if args.tensorboard==True:
        from tensorflow.keras.callbacks import TensorBoard
        cur_logdir='/'.join([args.tensorboard_logdir,model_output_path_string.split('/')[-1]+'.tb'])
        if not os.path.exists(cur_logdir):
                os.makedirs(cur_logdir)
        tensorboard_visualizer=TensorBoard(log_dir=cur_logdir, histogram_freq=0, batch_size=500, write_graph=True, write_grads=False, write_images=False, embeddings_freq=0, embeddings_layer_names=None, embeddings_metadata=None)
        cur_callbacks.append(tensorboard_visualizer)
    model.fit(train_gen,
              validation_data=valid_gen,
              steps_per_epoch=args.num_train/args.batch_size,
              validation_steps=args.num_valid/args.batch_size,
              epochs=args.epochs,
              verbose=1,
              use_multiprocessing=args.use_multiprocessing,
              workers=args.threads,
              max_queue_size=args.max_queue_size,
              callbacks=cur_callbacks,
              shuffle=True)
    print('fit_generator complete') 
    model.save_weights(model_output_path_weights_name)
    print('weights saved') 
    architecture_string=model.to_json()
    with open(model_output_path_arch_name,'w') as outf:
        outf.write(architecture_string)
    print('saved model architecture')
    #sync to s3 if needed
    if args.model_prefix.startswith('s3://'):
        #sync log, model hdf5, weight file, arch file
        upload_s3_file(args.model_prefix+'.hdf5',model_output_path_hdf5_name)
        upload_s3_file(args.model_prefix+'.arch',model_output_path_arch_name)
        upload_s3_file(args.model_prefix+'.log',model_output_path_logs_name)
        upload_s3_file(args.model_prefix+'.weights',model_output_path_weights_name)                
    print("complete!!")
    
def initializer_generators_hdf5(args):
    #get upsampling parameters
    if args.train_chroms is not None:
        train_chroms=args.train_chroms
    else: 
        train_chroms=get_chroms(args,split='train')
    index_train_path, index_valid_path, input_train_path, input_valid_path, output_train_path, output_valid_path=get_paths(args)
    train_generator=DataGenerator(index_path=index_train_path,
                                  input_path=input_train_path,
                                  output_path=output_train_path,
                                  index_tasks=args.index_tasks,
                                  num_inputs=args.num_inputs,
                                  num_outputs=args.num_outputs,
                                  ref_fasta=args.ref_fasta,
                                  batch_size=args.batch_size,
                                  add_revcomp=args.revcomp,
                                  chroms_to_use=train_chroms,
                                  get_w1_w0=args.weighted,
                                  upsample_thresh_list=args.upsample_thresh_list_train,
                                  upsample_ratio_list=args.upsample_ratio_list_train,
                                  tasks=args.hdf5_tasks,
                                  expand_dims=args.expand_dims,
                                  shuffle=args.shuffle_epoch_end)    
    print("generated training data generator!")

    if args.validation_chroms is not None:
        valid_chroms=args.validation_chroms
    else: 
        valid_chroms=get_chroms(args,split='valid')
    valid_generator=DataGenerator(index_path=index_train_path,
                                  input_path=input_train_path,
                                  output_path=output_train_path,
                                  index_tasks=args.index_tasks,
                                  num_inputs=args.num_inputs,
                                  num_outputs=args.num_outputs,
                                  ref_fasta=args.ref_fasta,
                                  batch_size=args.batch_size,
                                  add_revcomp=args.revcomp,                        
                                  upsample_thresh_list=args.upsample_thresh_list_eval,
                                  upsample_ratio_list=args.upsample_ratio_list_eval,
                                  chroms_to_use=valid_chroms,
                                  expand_dims=args.expand_dims,
                                  tasks=args.hdf5_tasks,
                                  shuffle=args.shuffle_epoch_end)
    print("generated validation data generator!")
    return train_generator, valid_generator 

def initialize_generators_tiledb(args):

    if args.upsample_ratio_list_train is not None:
        upsample_ratio_train=args.upsample_ratio_list_train[0]
        print("warning! only a single ratio for upsampling supported for tiledb as of now")
    else:
        upsample_ratio_train=None
    if args.upsample_ratio_list_eval is not None:
        upsample_ratio_eval=args.upsample_ratio_list_eval[0]
        print("warning! only a single ratio for upsampling supported for tiledb as of now")
    else:
        upsample_ratio_eval=None

    #if bed regions are provided, generate splits
    if args.bed_regions is not None:
        assert args.bed_regions_train is None
        assert args.bed_regions_validate is None
        assert args.genome is not None
        assert args.fold is not None
        bed_regions=pd.read_csv(args.bed_regions,header=None,sep='\t')
        bed_regions_train=get_bed_regions_for_fold_split(bed_regions,args.genome,args.fold,'train')
        bed_regions_validate=get_bed_regions_for_fold_split(bed_regions,args.genome,args.fold,'valid')
    elif (args.bed_regions_train is not None) or (args.bed_regions_validate is not None):
        #make sure that both training and validation bed files are provided 
        assert args.bed_regions_validate is not None
        assert args.bed_regions_train is not None        
        bed_regions_train=args.bed_regions_train
        bed_regions_validate=args.bed_regions_validate
    else:
        bed_regions_train=args.bed_regions_train #should be None
        bed_regions_validate=args.bed_regions_validate #should be None
        
    import tiledb
    tdb_config=get_default_config() 
    tdb_ctx=tiledb.Ctx(config=tdb_config)
    train_chroms=get_chroms(args,split='train')
    train_generator=TiledbGenerator(chroms=train_chroms,
                                    ref_fasta=args.ref_fasta,
                                    shuffle_epoch_start=True,
                                    shuffle_epoch_end=args.shuffle_epoch_end,
                                    batch_size=args.batch_size,
                                    tdb_array=args.tdb_array,
                                    tdb_partition_attribute_for_upsample=args.tdb_partition_attribute_for_upsample,
                                    tdb_partition_thresh_for_upsample=args.tdb_partition_thresh_for_upsample,
                                    tdb_partition_datasets_for_upsample=args.tdb_partition_datasets_for_upsample,
                                    tdb_input_source_attribute=args.tdb_input_source_attribute,
                                    tdb_input_flank=args.tdb_input_flank,
                                    tdb_input_min=args.tdb_input_min,
                                    tdb_input_max=args.tdb_input_max,
                                    tdb_input_aggregation=args.tdb_input_aggregation,
                                    tdb_input_transformation=args.tdb_input_transformation,
                                    pseudocount=args.tdb_transformation_pseudocount,
                                    tdb_output_source_attribute=args.tdb_output_source_attribute,
                                    tdb_output_flank=args.tdb_output_flank,
                                    tdb_output_min=args.tdb_output_min,
                                    tdb_output_max=args.tdb_output_max,
                                    tdb_output_aggregation=args.tdb_output_aggregation,
                                    tdb_output_transformation=args.tdb_output_transformation,
                                    tdb_ambig_attribute=args.tdb_ambig_attribute,
                                    tdb_input_datasets=args.tdb_input_datasets,
                                    tdb_output_datasets=args.tdb_output_datasets,
                                    upsample_ratio=upsample_ratio_train,
                                    num_inputs=args.num_inputs,
                                    num_outputs=args.num_outputs,
                                    add_revcomp=args.revcomp,
                                    tdb_config=tdb_config,
                                    tdb_ctx=tdb_ctx,
                                    num_threads=args.upsample_threads,
                                    bed_regions=bed_regions_train,
                                    bed_regions_center=args.bed_regions_center,
                                    bed_regions_jitter=args.bed_regions_jitter)
    
    print("generated training data generator!")
    valid_chroms=get_chroms(args,split='valid')
    valid_generator=TiledbGenerator(chroms=valid_chroms,
                                    ref_fasta=args.ref_fasta,
                                    shuffle_epoch_start=True,
                                    shuffle_epoch_end=args.shuffle_epoch_end,
                                    batch_size=args.batch_size,
                                    tdb_array=args.tdb_array,
                                    tdb_partition_attribute_for_upsample=args.tdb_partition_attribute_for_upsample,
                                    tdb_partition_thresh_for_upsample=args.tdb_partition_thresh_for_upsample,
                                    tdb_partition_datasets_for_upsample=args.tdb_partition_datasets_for_upsample,
                                    tdb_input_source_attribute=args.tdb_input_source_attribute,
                                    tdb_input_flank=args.tdb_input_flank,
                                    tdb_input_min=args.tdb_input_min,
                                    tdb_input_max=args.tdb_input_max,
                                    tdb_input_aggregation=args.tdb_input_aggregation,
                                    tdb_input_transformation=args.tdb_input_transformation,
                                    pseudocount=args.tdb_transformation_pseudocount,
                                    tdb_output_source_attribute=args.tdb_output_source_attribute,
                                    tdb_output_flank=args.tdb_output_flank,
                                    tdb_output_min=args.tdb_output_min,
                                    tdb_output_max=args.tdb_output_max,
                                    tdb_output_aggregation=args.tdb_output_aggregation,
                                    tdb_output_transformation=args.tdb_output_transformation,
                                    tdb_ambig_attribute=args.tdb_ambig_attribute,
                                    tdb_input_datasets=args.tdb_input_datasets,
                                    tdb_output_datasets=args.tdb_output_datasets,
                                    upsample_ratio=upsample_ratio_eval,
                                    num_inputs=args.num_inputs,
                                    num_outputs=args.num_outputs,
                                    add_revcomp=args.revcomp,
                                    tdb_config=tdb_config,
                                    tdb_ctx=tdb_ctx,
                                    num_threads=args.upsample_threads,
                                    bed_regions=bed_regions_validate,
                                    bed_regions_center=args.bed_regions_center,
                                    bed_regions_jitter=args.bed_regions_jitter)
    
    print("generated validation data generator")
    return train_generator, valid_generator

def get_paths(args):
    if args.index_train_path is None:
        index_train_path=args.index_data_path
    else:
        index_train_path=args.index_train_path        
    if args.input_train_path is None:
        input_train_path=args.input_data_path
    else:
        input_train_path=args.input_train_path        
    if args.output_train_path is None:
        output_train_path=args.output_data_path
    else:
        output_train_path=args.output_train_path        
    if args.index_valid_path is None:
        index_valid_path=args.index_data_path
    else:
        index_valid_path=args.index_valid_path
    if args.input_valid_path is None:
        input_valid_path=args.input_data_path
    else:
        input_valid_path=args.input_valid_path
    if args.output_valid_path is None:
        output_valid_path=args.output_data_path
    else:
        output_valid_path=args.output_valid_path        
    return index_train_path, index_valid_path, input_train_path, input_valid_path, output_train_path, output_valid_path 

def initialize_generators(args):    
    #data is being read in from tiledb for training
    print(args)
    if args.tdb_array is not None:
        return initialize_generators_tiledb(args)
    else:
        return initializer_generators_hdf5(args)
    
    
def train(args):
    if type(args)==type({}):
        args=config.args_object_from_args_dict(args)


    #create the generators
    train_generator,valid_generator=initialize_generators(args)
    
    w1,w0=get_w1_w0_training(args,train_generator)
    args.w1=w1
    args.w0=w0
    model=get_model(args)
    fit_and_evaluate(model,train_generator,
                     valid_generator,args)

    #remove any temporary s3 files
    print("running cleanup!") 
    run_cleanup()
    
def main():
    gc.freeze()
    args=parse_args()
    train(args)
    print("Exiting!")

if __name__=="__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import time
#graceful shutdown
import psutil
import signal 
import os
#multithreading
#from concurrent.futures import ProcessPoolExecutor, as_completed
from multiprocessing import Pool,Process, Queue 
import warnings
import numpy as np
import pysam
import pandas as pd
import tensorflow as tf 
from kerasAC.activations import softMaxAxis1
from .calibrate import * 
from .generators.basic_generator import *
from .generators.tiledb_predict_generator import *
from .tiledb_config import *
from .s3_sync import *
from .get_model import *
from .splits import *
from kerasAC.config import args_object_from_args_dict
from kerasAC.performance_metrics import *
from kerasAC.custom_losses import *
from kerasAC.metrics import recall, specificity, fpr, fnr, precision, f1
import argparse
import yaml 
import h5py 
import pickle
import numpy as np 
import tensorflow.keras as keras 
from keras.losses import *
from keras.models import Model
from keras.utils import multi_gpu_model
from kerasAC.custom_losses import *
from abstention.calibration import PlattScaling, IsotonicRegression 
import random
import pdb 

def parse_args():
    parser=argparse.ArgumentParser(description='Provide model files  & a dataset, get model predictions')
    parser.add_argument('--ref_fasta')

    tiledbgroup=parser.add_argument_group("tiledb")
    tiledbgroup.add_argument("--tdb_array",help="name of tdb array to use")
    
    tiledbgroup.add_argument("--tdb_output_source_attribute",nargs="+",default=None,help="tiledb attribute for use in label generation i.e. fc_bigwig")
    tiledbgroup.add_argument("--tdb_output_datasets",nargs="+",default=None,help="dataset column from db_ingest; comma separated across channels; space separated across outputs")    
    tiledbgroup.add_argument("--tdb_output_min",nargs="*", default=None)
    tiledbgroup.add_argument("--tdb_output_max",nargs="*", default=None)        
    tiledbgroup.add_argument("--tdb_output_flank",nargs="+",help="flank around bin center to use in generating outputs")
    tiledbgroup.add_argument("--tdb_output_aggregation",nargs="+",default=None,help="method for output aggreagtion; one of None, 'avg','max'")
    tiledbgroup.add_argument("--tdb_output_transformation",nargs="+",default=None,help="method for output transformation; one of None, 'log','log10','asinh'")

    tiledbgroup.add_argument("--tdb_input_source_attribute",nargs="+",help="attribute to use for generating model input, or 'seq' for one-hot-encoded sequence")
    tiledbgroup.add_argument("--tdb_input_min",nargs="*", default=None)
    tiledbgroup.add_argument("--tdb_input_max",nargs="*", default=None)    
    tiledbgroup.add_argument("--tdb_input_datasets",nargs="+",default=None,help="dataset column from db_ingest; comma separated across channels; space separated across inputs")
    tiledbgroup.add_argument("--tdb_input_flank",nargs="+",help="length of sequence around bin center to use for input")
    tiledbgroup.add_argument("--tdb_input_aggregation",nargs="+",default=None,help="method for input aggregation; one of 'None','avg','max'")
    tiledbgroup.add_argument("--tdb_input_transformation",nargs="+",default=None,help="method for input transformation; one of None, 'log','log10','asinh'")
    tiledbgroup.add_argument("--tdb_transformation_pseudocount",type=float,default=1)
    
    tiledbgroup.add_argument("--tdb_partition_attribute_for_upsample",default="idr_peak",help="tiledb attribute to use for upsampling, i.e. idr_peak")
    tiledbgroup.add_argument("--tdb_partition_thresh_for_upsample",type=float,default=1,help="values >= partition_thresh_for_upsample within the partition_attribute_for_upsample will be upsampled during training")
    tiledbgroup.add_argument("--tdb_partition_datasets_for_upsample",nargs="+")
    
    tiledbgroup.add_argument("--upsample_ratio_list_predict",type=float,nargs="*")
    tiledbgroup.add_argument("--tdb_ambig_attribute",default=None,help="attribute indicating ambiguous regions to not train on")
    tiledbgroup.add_argument("--chrom_sizes",default=None,help="chromsizes file for use with tiledb generator")
    tiledbgroup.add_argument("--tiledb_stride",type=int,default=1)
    tiledbgroup.add_argument("--upsample_threads",type=int,default=1)
    tiledbgroup.add_argument("--revcomp",action="store_true")
    input_filtering_params=parser.add_argument_group("input_filtering_params")    
    input_filtering_params.add_argument('--predict_chroms',nargs="*",default=None)
    input_filtering_params.add_argument("--genome",default=None)
    input_filtering_params.add_argument("--fold",type=int,default=None)
    input_filtering_params.add_argument("--bed_regions",default=None) 
    input_filtering_params.add_argument('--bed_regions_center', choices=['random','center','summit'],default=None)
    input_filtering_params.add_argument("--datasets",nargs="*",default=None)
    input_filtering_params.add_argument("--dataset_indices",nargs="*",default=None)
    
    output_params=parser.add_argument_group("output_params")
    output_params.add_argument('--predictions_and_labels_hdf5',help='name of hdf5 to save predictions',default=None)
    calibration_params=parser.add_argument_group("calibration_params")
    calibration_params.add_argument("--calibrate_classification",action="store_true",default=False)
    calibration_params.add_argument("--calibrate_regression",action="store_true",default=False)        
    
    weight_params=parser.add_argument_group("weight_params")
    weight_params.add_argument('--w1',nargs="*",type=float)
    weight_params.add_argument('--w0',nargs="*",type=float)
    weight_params.add_argument("--w1_w0_file",default=None)


    model_params=parser.add_argument_group("model_params")
    model_params.add_argument('--load_model_hdf5',help='hdf5 file that stores the model')
    model_params.add_argument('--weights',help='weights file for the model')
    model_params.add_argument('--yaml',help='yaml file for the model')
    model_params.add_argument('--json',help='json file for the model')
    model_params.add_argument('--functional',default=False,help='use this flag if your model is a functional model',action="store_true")
    model_params.add_argument('--squeeze_input_for_gru',action='store_true')
    model_params.add_argument("--num_inputs",type=int)
    model_params.add_argument("--num_outputs",type=int)
    model_params.add_argument("--num_gpus",type=int,default=1)
    
    snp_params=parser.add_argument_group("snp_params")
    snp_params.add_argument('--background_freqs',default=None)
    snp_params.add_argument('--flank',default=500,type=int)
    snp_params.add_argument('--mask',default=10,type=int)
    snp_params.add_argument('--ref_col',type=int,default=None)
    snp_params.add_argument('--alt_col',type=int,default=None)

    parser.add_argument('--batch_size',type=int,help='batch size to use to make model predictions',default=50)
    return parser.parse_args()

def write_predictions(args):
    '''
    keys lab_i, pred_i indicate the labels and predictions matrices for loss i. 
    The tasks are encoded in the final axis. i.e. profile labels matrix would be (N,L,T), where N = number of samples, L= length of profile, T = number of tasks.
    '''
    try:
        if args.predictions_and_labels_hdf5.startswith('s3://'):
            #use a local version of the file and upload to s3 when finished
            bucket,filename=s3_string_parse(args.predictions_and_labels_hdf5)
            out_predictions_prefix=filename.split('/')[-1]+".predictions"
        else: 
            out_predictions_prefix=args.predictions_and_labels_hdf5+".predictions"
        #create output h5py file
        outf=h5py.File(out_predictions_prefix,'w')
        first=True
        while True:
            cur_out=queue.get()
            if type(cur_out)==str:
                if cur_out=="FINISHED":
                    outf.close()
                    return
            coords=cur_out[0]
            labels=cur_out[1]
            pred=cur_out[2]
            num_datasets=len(pred)
            if first is True:
                #create the predictions & labels datasets
                for i in range(num_datasets):
                    cur_pred_shape=pred[i].shape
                    if len(cur_pred_shape)==1:
                        maxshape=(None,)
                    else:
                        maxshape=(None,)+cur_pred_shape[1::]
                        
                    outf.create_dataset('pred_'+str(i),data=pred[i],maxshape=maxshape)
                    outf.create_dataset('lab_'+str(i),data=labels[i],maxshape=maxshape)
                #create the coord dataset
                coord_shape=coords.shape
                if len(coord_shape)==1:
                    maxshape=(None,)
                else:
                    maxshape=(None,)+coord_shape[1::]
                outf.create_dataset('coords',data=coords,maxshape=maxshape)
                

                first=False
            else:
                #append to hdf5
                
                #append coords
                curshape=len(outf['coords'])
                batchshape=coords.shape[0]
                newshape=curshape+batchshape
                outf['coords'].resize(newshape,axis=0)
                outf['coords'][-1*batchshape::]=coords
                
                #append predictions
                for i in range(num_datasets): 
                    curshape=len(outf['pred_'+str(i)])
                    batchshape=pred[i].shape[0]
                    newshape=curshape+batchshape
                    outf['pred_'+str(i)].resize(newshape,axis=0)
                    outf['pred_'+str(i)][-1*batchshape::]=pred[i]                
                    outf['lab_'+str(i)].resize(newshape,axis=0)
                    outf['lab_'+str(i)][-1*batchshape::]=labels[i]                
    except KeyboardInterrupt:
        #shutdown the pool
        # Kill remaining child processes
        kill_child_processes(os.getpid())
        raise 
    except Exception as e:
        print(e)
        #shutdown the pool
        # Kill remaining child processes
        kill_child_processes(os.getpid())
        raise e


def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

def kill_child_processes(parent_pid, sig=signal.SIGTERM):
    try:
        parent = psutil.Process(parent_pid)
    except psutil.NoSuchProcess:
        return
    children = parent.children(recursive=True)
    for process in children:
        process.send_signal(sig)
        
def get_batch_wrapper(idx):
    X,y,coords=test_generator[idx]
    if type(y) is not list:
        y=[y]
    try:
        y=[i.squeeze(axis=-1) for i in y]
    except:
        pass
    if type(X) is not list:
        X=[X]
    return X,y,coords
    #represent coords w/ string, MultiIndex not supported in table append mode
    #set the column names for the MultiIndex
    #coords=pd.MultiIndex.from_tuples(coords,names=['CHR','CENTER','STRAND'])
    #y=[pd.DataFrame(i,index=coords) for i in y]
    #return X,y,coords


def get_tiledb_predict_generator(args):
    global test_generator
    if args.upsample_ratio_list_predict is not None:
        upsample_ratio_predict=args.upsample_ratio_list_predict[0]
        print("warning! only a single ratio for upsampling supported for tiledb as of now")
    else:
        upsample_ratio_predict=None
    import tiledb
    tdb_config=get_default_config() 
    tdb_ctx=tiledb.Ctx(config=tdb_config)
    #you can only specify one (or neither) or args.fold or args.predict chroms 
    assert (args.fold is None) or (args.predict_chroms is None)
    if args.fold is None:
        predict_chroms=args.predict_chroms
    else:
        predict_chroms=get_chroms(args,split='test')
    bed_regions=pd.read_csv(args.bed_regions,header=None,sep='\t')
    bed_regions_test=get_bed_regions_for_fold_split(bed_regions,args.genome,args.fold,'test')
    test_generator=TiledbPredictGenerator(ref_fasta=args.ref_fasta,
                                          batch_size=args.batch_size,
                                          tdb_array=args.tdb_array,
                                          tdb_partition_attribute_for_upsample=args.tdb_partition_attribute_for_upsample,
                                          tdb_partition_thresh_for_upsample=args.tdb_partition_thresh_for_upsample,
                                          tdb_partition_datasets_for_upsample=args.tdb_partition_datasets_for_upsample,
                                          upsample_ratio=upsample_ratio_predict,
                                          num_threads=args.upsample_threads,
                                          tdb_ambig_attribute=args.tdb_ambig_attribute,
                                          tdb_input_source_attribute=args.tdb_input_source_attribute,
                                          tdb_input_flank=args.tdb_input_flank,
                                          tdb_input_min=args.tdb_input_min,
                                          tdb_input_max=args.tdb_input_max,
                                          tdb_output_source_attribute=args.tdb_output_source_attribute,
                                          tdb_output_flank=args.tdb_output_flank,
                                          tdb_output_min=args.tdb_output_min,
                                          tdb_output_max=args.tdb_output_max,
                                          num_inputs=args.num_inputs,
                                          num_outputs=args.num_outputs,
                                          tdb_input_aggregation=args.tdb_input_aggregation,
                                          tdb_input_transformation=args.tdb_input_transformation,
                                          pseudocount=args.tdb_transformation_pseudocount,
                                          tdb_output_aggregation=args.tdb_output_aggregation,
                                          tdb_output_transformation=args.tdb_output_transformation,                                          
                                          tiledb_stride=args.tiledb_stride,
                                          chrom_sizes=args.chrom_sizes,
                                          chroms=predict_chroms,
                                          tdb_input_datasets=args.tdb_input_datasets,
                                          tdb_output_datasets=args.tdb_output_datasets,
                                          tdb_config=tdb_config,
                                          tdb_ctx=tdb_ctx,
                                          bed_regions=bed_regions_test,
                                          bed_regions_center=args.bed_regions_center,
                                          add_revcomp=args.revcomp)
    print("created TiledbPredictGenerator")    
    return test_generator 

def predict_on_batch_wrapper(args,model,test_generator):
    num_batches=len(test_generator)
    for idx in range(num_batches):
        if idx%100==0:
            print(str(idx)+'/'+str(num_batches))
        X,y,coords=get_batch_wrapper(idx)
        #get the model predictions            
        preds=model.predict_on_batch(X)
        if type(preds) is not list:
            preds=[preds]
        try:
            preds=[i.squeeze(axis=-1) for i in preds]
        except:
            pass 
        queue.put((coords,y,preds))
    print("finished with tiledb predictions!")
    queue.put("FINISHED")
    queue.close() 
    return

def get_model_layer_functor(model,target_layer_idx):
    from keras import backend as K
    inp=model.input
    outputs=model.layers[target_layer_idx].output
    functor=K.function([inp], [outputs])
    return functor 

def get_layer_outputs(functor,X):
    return functor([X])

def predict(args):
    if type(args)==type({}):
        args=args_object_from_args_dict(args) 
    global queue
    
    queue=Queue()
    
    
    writer=Process(target=write_predictions,args=([args]))
    writer.start()
    
    #get the generator
    test_generator=get_tiledb_predict_generator(args) 
    
    #get the model
    #if calibration is to be done, get the preactivation model 
    model=get_model(args)
    perform_calibration=args.calibrate_classification or args.calibrate_regression
    if perform_calibration==True:
        if args.calibrate_classification==True:
            print("getting logits")
            model=Model(inputs=model.input,
                               outputs=model.layers[-2].output)
        elif args.calibrate_regression==True:
            print("getting pre-relu outputs (preacts)")
            model=Model(inputs=model.input,
                        outputs=model.layers[-1].output)
            
    #call the predict_on_batch_wrapper
    predict_on_batch_wrapper(args,model,test_generator)

    #drain the queue
    try:
        while not queue.empty():
            print("draining the prediction Queue")
            time.sleep(2)
    except Exception as e:
        print(e)
    
    print("joining prediction writer") 
    writer.join()
    
    #sync files to s3 if needed
    if args.predictions_and_labels_hdf5.startswith('s3://'):
        #use a local version of the file and upload to s3 when finished
        out_predictions_prefix=args.predictions_and_labels_hdf5+".predictions"
        #upload outputs for all tasks
        import glob
        to_upload=[]
        for f in glob.glob(out_predictions_prefix+"*"):
            s3_path='/'.join(out_predictions_prefix.split('/')[0:-1]+[f.split('/')[-1]])
            print(f+'-->'+s3_path)
            upload_s3_file(s3_path,f)
        
    #perform calibration, if specified
    if perform_calibration is True:
        print("calibrating")
        calibrate(args)

    #clean up any s3 artifacts:
    run_cleanup()
    
def main():
    args=parse_args()
    predict(args)


if __name__=="__main__":
    main()
    



import sys, os
import math 
import os.path
from collections import OrderedDict
import argparse
import numpy as np 
import tiledb
import random 
ltrdict = {'a':[1,0,0,0],
           'c':[0,1,0,0],
           'g':[0,0,1,0],
           't':[0,0,0,1],
           'n':[0,0,0,0],
           'A':[1,0,0,0],
           'C':[0,1,0,0],
           'G':[0,0,1,0],
           'T':[0,0,0,1],
           'N':[0,0,0,0]}
reverse_dict={}
for key in ltrdict:
    cur_val=ltrdict[key]
    reverse_dict[tuple(cur_val)]=key

#def one_hot_encode(seqs):
#    return np.array([[ltrdict.get(x,[0,0,0,0]) for x in seq] for seq in seqs])

def one_hot_encode(seqs):
    """
    Converts a list of DNA ("ACGT") sequences to one-hot encodings, where the
    position of 1s is ordered alphabetically by "ACGT". `seqs` must be a list
    of N strings, where every string is the same length L. Returns an N x L x 4
    NumPy array of one-hot encodings, in the same order as the input sequences.
    All bases will be converted to upper-case prior to performing the encoding.
    Any bases that are not "ACGT" will be given an encoding of all 0s.
    """
    seq_len = len(seqs[0])
    assert np.all(np.array([len(s) for s in seqs]) == seq_len)

    # Join all sequences together into one long string, all uppercase
    seq_concat = "".join(seqs).upper()

    one_hot_map = np.identity(5)[:, :-1]

    # Convert string into array of ASCII character codes;
    base_vals = np.frombuffer(bytearray(seq_concat, "utf8"), dtype=np.int8)

    # Anything that's not an A, C, G, or T gets assigned a higher code
    base_vals[~np.isin(base_vals, np.array([65, 67, 71, 84]))] = 85

    # Convert the codes into indices in [0, 4], in ascending order by code
    _, base_inds = np.unique(base_vals, return_inverse=True)

    # Get the one-hot encoding for those indices, and reshape back to separate
    return one_hot_map[base_inds].reshape((len(seqs), seq_len, 4))


def one_hot_to_dna(one_hot):
    """
    Converts a one-hot encoding into a list of DNA ("ACGT") sequences, where the
    position of 1s is ordered alphabetically by "ACGT". `one_hot` must be an
    N x L x 4 array of one-hot encodings. Returns a lits of N "ACGT" strings,
    each of length L, in the same order as the input array. The returned
    sequences will only consist of letters "A", "C", "G", "T", or "N" (all
    upper-case). Any encodings that are all 0s will be translated to "N".
    """
    bases = np.array(["A", "C", "G", "T", "N"])
    # Create N x L array of all 5s
    one_hot_inds = np.tile(one_hot.shape[2], one_hot.shape[:2])

    # Get indices of where the 1s are
    batch_inds, seq_inds, base_inds = np.where(one_hot)

    # In each of the locations in the N x L array, fill in the location of the 1
    one_hot_inds[batch_inds, seq_inds] = base_inds

    # Fetch the corresponding base for each position using indexing
    seq_array = bases[one_hot_inds]
    return ["".join(seq) for seq in seq_array]

def dinuc_shuffle(seq):
    #get list of dinucleotides
    nucs=[]
    for i in range(0,len(seq),2):
        nucs.append(seq[i:i+2])
    #generate a random permutation
    random.shuffle(nucs)
    return ''.join(nucs) 


def revcomp(seq):
    seq=seq[::-1].upper()
    comp_dict=dict()
    comp_dict['A']='T'
    comp_dict['T']='A'
    comp_dict['C']='G'
    comp_dict['G']='C'
    rc=[]
    for base in seq:
        if base in comp_dict:
            rc.append(comp_dict[base])
        else:
            rc.append(base)
    return ''.join(rc)






class DefaultOrderedDictWrapper(object):
    def __init__(self, factory):
        self.ordered_dict = OrderedDict()
        assert hasattr(factory, '__call__')
        self.factory = factory

    def __getitem__(self, key):
        if key not in self.ordered_dict:
            self.ordered_dict[key] = self.factory() 
        return self.ordered_dict[key]

def enum(**enums):
    class Enum(object):
        pass
    to_return = Enum
    for key,val in enums.items():
        if hasattr(val, '__call__'): 
            setattr(to_return, key, staticmethod(val))
        else:
            setattr(to_return, key, val)
    to_return.vals = [x for x in enums.values()]
    to_return.the_dict = enums
    return to_return


def combine_enums(*enums):
    new_enum_dict = OrderedDict()
    for an_enum in enums:
        new_enum_dict.update(an_enum.the_dict)
    return enum(**new_enum_dict)


    
    
def coords_to_tdb_indices(coords,tdb_instance):
    '''
    coords is a tuple (chrom, start, stop)
    '''
    num_chroms=tdb_instance.meta['num_chroms']
    for i in range(num_chroms):
        if tdb_instance.meta['chrom_'+str(i)]==coords[0]:
            chrom_offset=tdb_instance.meta['offset_'+str(i)]
            tdb_index_start=chrom_offset+coords[1]
            tdb_index_end=chrom_offset+coords[2]
            return (tdb_index_start,tdb_index_end)
    raise Exception("chrom name:"+str(coords[0])+" not found in tdb array")


def tdb_indices_to_coords(indices,tdb_instance):
    '''
    indices is a list of tdb indices     
    '''
    pass


def transform_data_type(inputs,num_inputs):
    if inputs is None:
        inputs=[None]*num_inputs
    elif inputs is []:
        inputs=[None]*num_inputs
    else:
        assert(len(inputs)==num_inputs)
        transformed=[]
        for i in range(num_inputs):
            transformed.append([]) 
            cur_inputs=inputs[i].split(',')
            for j in cur_inputs: 
                if str(j).lower()=="none":
                    transformed[i].append(None)
                else:
                    transformed[i].append(float(j))
    return transformed


def transform_data_type_min(inputs,num_inputs):
    if inputs is None:
        transformed=[-math.inf]*num_inputs
    elif inputs is []:
        transformed=[-math.inf]*num_inputs
    else:
        assert(len(inputs)==num_inputs)
        transformed=[]
        for i in range(num_inputs):
            transformed.append([]) 
            cur_inputs=inputs[i].split(',')
            for j in cur_inputs: 
                if str(j).lower()=="none":
                    transformed[i].append(-math.inf)
                else:
                    transformed[i].append(float(j))
    return transformed

def transform_data_type_max(inputs,num_inputs):
    if inputs is None:
        transformed=[math.inf]*num_inputs
    elif inputs is []:
        transformed=[math.inf]*num_inputs
    else:
        assert(len(inputs)==num_inputs)
        transformed=[]
        for i in range(num_inputs):
            transformed.append([]) 
            cur_inputs=inputs[i].split(',')
            for j in cur_inputs: 
                if str(j).lower()=="none":
                    transformed[i].append(math.inf)
                else:
                    transformed[i].append(float(j))
    return transformed




import numpy as np ;
from keras.constraints import max_norm
from kerasAC.metrics import *
from kerasAC.custom_losses import get_weighted_binary_crossentropy, get_ambig_binary_crossentropy
from kerasAC.metrics import recall, specificity, fpr, fnr, precision, f1
import keras;
from keras.models import Sequential
from keras.layers.core import Dropout, Reshape, Dense, Activation, Flatten
from keras.layers.convolutional import Conv2D, MaxPooling2D
from keras.optimizers import Adadelta, SGD, RMSprop;
import keras.losses;
from keras.constraints import maxnorm;
from keras.layers.normalization import BatchNormalization
from keras.regularizers import l1, l2    
from keras import backend as K
K.set_image_data_format('channels_last')
print(K.image_data_format())

def getModelGivenModelOptionsAndWeightInits(args):
    #read in the args
    seed=args.seed
    ntasks=args.ntasks
    w0=args.w0
    w1=args.w1
    
    np.random.seed(seed)
    model=Sequential()
    model.add(Conv2D(filters=50,kernel_size=(1,15),padding="same", kernel_constraint=max_norm(7.0,axis=-1),input_shape=(1,1000,4)))
    model.add(BatchNormalization(axis=-1))
    model.add(Activation('relu'))   
    model.add(Conv2D(filters=50,kernel_size=(1,15),padding="same"))
    model.add(BatchNormalization(axis=-1))
    model.add(Activation('relu'))    
    model.add(Conv2D(filters=50,kernel_size=(1,13),padding="same"))
    model.add(BatchNormalization(axis=-1))
    model.add(Activation('relu'))    
    model.add(MaxPooling2D(pool_size=(1,40)))    
    model.add(Flatten())
    model.add(Dense(50))
    model.add(BatchNormalization(axis=-1))
    model.add(Activation('relu'))
    model.add(Dropout(0.2))    
    model.add(Dense(ntasks))
    model.add(Activation("sigmoid"))
    adam = keras.optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
    print("compiling!")
    if w0!=None:
        loss=get_weighted_binary_crossentropy(w0_weights=w0,w1_weights=w1)
    else:
        loss=get_ambig_binary_crossentropy() 
    model.compile(optimizer=adam,
                  loss=loss,
                  metrics=[recall, specificity, fpr, fnr, precision, f1])
    return model



import numpy as np ;
from kerasAC.metrics import * 

def getModelGivenModelOptionsAndWeightInits(args):
    #read in the arguments
    seed=args.seed
    ntasks=args.ntasks
    w0=args.w0
    w1=args.w1
    init_weights=args.init_weights
    
    np.random.seed(seed)
    import keras;
    from keras.layers import (
        Activation, AveragePooling1D, BatchNormalization,
        Conv1D, Conv2D, Dense, Dropout, Flatten, Input,
        MaxPooling1D, MaxPooling2D, Reshape,
        PReLU, Add
    )
    from keras.models import Model
    #from keras.layers.core import Dropout, Reshape, Dense, Activation, Flatten
    #from keras.layers.convolutional import Conv2D, MaxPooling2D
    from keras.optimizers import Adadelta, SGD, RMSprop;
    import keras.losses;
    from keras.constraints import maxnorm;
    from keras.layers.normalization import BatchNormalization
    from keras.regularizers import l1, l2    
    from keras import backend as K
    #K.set_image_data_format('channels_last')
    print(K.image_data_format())

    import collections
    model_inputs = ["data/genome_data_dir"]
    shapes = {'data/genome_data_dir': [1000, 4]}
    keras_inputs = collections.OrderedDict([(name, Input(shape=shapes[name], name=name)) for name in model_inputs])
    inputs = keras_inputs
    num_tasks = ntasks
    seq_preds = inputs["data/genome_data_dir"]
    num_filters = (48, 64, 100, 150, 300, 200, 200, 200, 200)
    conv_width = (3, 3, 3, 7, 7, 7, 3, 3, 7)
    batch_norm = True
    pool_width=(3, 4, 4)
    pool_stride=(3, 4, 4)
    fc_layer_sizes=(1000, 1000)
    dropout=(0.3, 0.3)
    final_dropout=0.0,
    trainable=1
    final_layer_name='tuned_i_score'
    j = 0
    for i, (nb_filter, nb_col) in enumerate(zip(num_filters, conv_width)):
        seq_preds = Conv1D(nb_filter, nb_col, kernel_initializer='he_normal', trainable = bool(trainable))(seq_preds)
        if batch_norm:
            seq_preds = BatchNormalization(trainable = bool(trainable))(seq_preds)
        seq_preds = Activation('relu', trainable = bool(trainable))(seq_preds)

        if(i == 4 or i == 7 or i == 8):
            seq_preds = MaxPooling1D(pool_width[j], pool_stride[j], trainable = bool(trainable))(seq_preds)
            j = j+1

    seq_preds = Flatten()(seq_preds)

    # fully connect, drop before fc layers
    for drop_rate, fc_layer_size in zip(dropout, fc_layer_sizes):
        seq_preds = Dense(fc_layer_size)(seq_preds)
        if batch_norm:
            seq_preds = BatchNormalization()(seq_preds)
        seq_preds = Activation('relu')(seq_preds)
        #seq_preds = Dropout(drop_rate)(seq_preds)
    #seq_preds = Dropout(final_dropout)(seq_preds)
    seq_preds = Dense(num_tasks, name=final_layer_name)(seq_preds)
    #seq_preds = Activation('relu')(seq_preds)
    random_weight_model = Model(inputs=list(keras_inputs.values()), outputs=seq_preds)

    model = random_weight_model


    if (init_weights!=None):
        #load the weight initializations
        model.load_weights(init_weights, by_name=True)

    adam = keras.optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
    print("compiling!")
    model.compile(optimizer=adam,loss='mse',metrics=[positive_accuracy,negative_accuracy,precision,recall])

    return model



import numpy as np ;
from keras.constraints import max_norm
from kerasAC.metrics import *
from kerasAC.custom_losses import get_weighted_binary_crossentropy, get_ambig_binary_crossentropy
from kerasAC.metrics import recall, specificity, fpr, fnr, precision, f1


def getModelGivenModelOptionsAndWeightInits(args):
    seed=args.seed
    w0=args.w0
    w1=args.w1
    ntasks=args.ntasks
    init_weights=args.init_weights
    
    np.random.seed(seed)
    import keras;
    from keras.models import Sequential
    from keras.layers.core import Dropout, Reshape, Dense, Activation, Flatten
    from keras.layers.convolutional import Conv2D, MaxPooling2D
    from keras.optimizers import Adadelta, SGD, RMSprop;
    import keras.losses;
    from keras.constraints import maxnorm;
    from keras.layers.normalization import BatchNormalization
    from keras.regularizers import l1, l2    
    from keras import backend as K
    K.set_image_data_format('channels_last')
    print(K.image_data_format())

    if (init_weights!=None):
        #load the weight initializations 
        data=np.load(init_weights);
        model=Sequential()
        model.add(Conv2D(filters=300,kernel_size=(1,19),input_shape=(1,1000,4),weights=[data['0.Conv/weights:0'],np.zeros(300,)],padding="same"))
        model.add(BatchNormalization(axis=-1,weights=[data['2.BatchNorm/gamma:0'],data['1.BatchNorm/beta:0'],np.zeros(300,),np.zeros(300,)]))
        model.add(Activation('relu'))
        model.add(MaxPooling2D(pool_size=(1,3)))
        model.add(Conv2D(filters=200,kernel_size=(1,11),weights=[data['3.Conv_1/weights:0'],np.zeros(200,)],padding="same"))
        model.add(BatchNormalization(axis=-1,weights=[data['5.BatchNorm_1/gamma:0'],data['4.BatchNorm_1/beta:0'],np.zeros(200,),np.zeros(200,)]))
        model.add(Activation('relu'))
        model.add(MaxPooling2D(pool_size=(1,4)))
        model.add(Conv2D(filters=200,kernel_size=(1,7),weights=[data['6.Conv_2/weights:0'],np.zeros(200,)],padding="same"))
        model.add(BatchNormalization(axis=-1,weights=[data['8.BatchNorm_2/gamma:0'],data['7.BatchNorm_2/beta:0'],np.zeros(200,),np.zeros(200,)]))
        model.add(Activation('relu'))
        model.add(MaxPooling2D(pool_size=(1,4)))

        model.add(Flatten())
        model.add(Dense(1000,weights=[data['9.fc0/fully_connected/weights:0'],np.zeros(1000,)]))
        model.add(BatchNormalization(axis=1,weights=[data['11.fc0/BatchNorm/gamma:0'],data['10.fc0/BatchNorm/beta:0'],np.zeros(1000,),np.zeros(1000,)]))
        model.add(Activation('relu'))
        model.add(Dropout(0.3))

        model.add(Dense(1000,weights=[data['12.fc1/fully_connected/weights:0'],np.zeros(1000,)]))
        model.add(BatchNormalization(axis=1,weights=[data['14.fc1/BatchNorm/gamma:0'],data['13.fc1/BatchNorm/beta:0'],np.zeros(1000,),np.zeros(1000,)]))
        model.add(Activation('relu'))
        model.add(Dropout(0.3))

        model.add(Dense(ntasks))
        model.add(Activation("sigmoid"))

    else:
        model=Sequential()
        model.add(Conv2D(filters=300,kernel_size=(1,19),input_shape=(1,1000,4)))
        model.add(BatchNormalization(axis=-1))
        model.add(Activation('relu'))
        model.add(MaxPooling2D(pool_size=(1,3)))

        model.add(Conv2D(filters=200,kernel_size=(1,11)))
        model.add(BatchNormalization(axis=-1))
        model.add(Activation('relu'))
        model.add(MaxPooling2D(pool_size=(1,4)))

        model.add(Conv2D(filters=200,kernel_size=(1,7)))
        model.add(BatchNormalization(axis=-1))
        model.add(Activation('relu'))
        model.add(MaxPooling2D(pool_size=(1,4)))

        model.add(Flatten())
        model.add(Dense(1000))
        model.add(BatchNormalization(axis=-1))
        model.add(Activation('relu'))
        model.add(Dropout(0.3))

        model.add(Dense(1000))
        model.add(BatchNormalization(axis=-1))
        model.add(Activation('relu'))
        model.add(Dropout(0.3))

        model.add(Dense(ntasks))
        model.add(Activation("sigmoid"))
        
    adam = keras.optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
    print("compiling!")
    if w0!=None:
        loss=get_weighted_binary_crossentropy(w0_weights=w0,w1_weights=w1)
    else:
        loss=ambig_binary_crossentropy 
    model.compile(optimizer=adam,
                  loss=loss,
                  metrics=[recall, specificity, fpr, fnr, precision, f1])
    return model



import tiledb
tdb_config_params={"sm.check_coord_dups":False,
                   "sm.check_coord_oob":False,
                   "sm.check_global_order":False,
                   "sm.num_writer_threads":40,
                   "sm.num_reader_threads":40,
                   "sm.num_async_threads":40,
                   "vfs.num_threads":40}

def get_default_config():
    tdb_config=tiledb.Config()
    tdb_config['vfs.s3.region']='us-west-1'
    tdb_config["sm.check_coord_dups"]="false"
    tdb_config["sm.check_coord_oob"]="false"
    tdb_config["sm.check_global_order"]="false"
    tdb_config["sm.num_reader_threads"]="40"
    tdb_config["sm.num_async_threads"]="40"
    tdb_config["vfs.num_threads"]="40"
    tdb_config["vfs.file.enable_filelocks"]="false"
    return tdb_config



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import multiprocessing as mp
import time
#graceful shutdown
import psutil
import signal 
import os

#multithreading
#from concurrent.futures import ProcessPoolExecutor, as_completed
#from multiprocessing import Pool,Process, Queue 

import warnings
import numpy as np
import pysam
import pandas as pd

import tensorflow as tf 
from kerasAC.activations import softMaxAxis1
from .calibrate import * 
from .generators.basic_generator import *
from .generators.tiledb_predict_generator import *
from .tiledb_config import *
from .s3_sync import *
from .splits import *
from .get_model import *
from .custom_losses import * 
from kerasAC.config import args_object_from_args_dict
from kerasAC.performance_metrics import *
from kerasAC.custom_losses import *
from kerasAC.metrics import recall, specificity, fpr, fnr, precision, f1
import argparse
import yaml 
import h5py 
import pickle
import numpy as np 
import tensorflow.keras as keras 
from keras.losses import *
from keras.models import Model
from keras.utils import multi_gpu_model
from kerasAC.custom_losses import *
from abstention.calibration import PlattScaling, IsotonicRegression 
import random
import pdb 

def parse_args():
    parser=argparse.ArgumentParser(description='Provide model files  & a dataset, get model predictions')
    input_data_path=parser.add_argument_group('input_data_path')
    input_data_path.add_argument("--index_data_path",default=None,help="seqdataloader output hdf5, or tsv file containing binned labels")
    input_data_path.add_argument("--index_tasks",nargs="*",default=None)    
    input_data_path.add_argument("--input_data_path",nargs="+",default=None,help="seq or path to seqdataloader hdf5")
    input_data_path.add_argument("--output_data_path",nargs="+",default=None,help="path to seqdataloader hdf5")
    
    input_data_path.add_argument('--variant_bed',default=None)
    input_data_path.add_argument('--ref_fasta')

    input_filtering_params=parser.add_argument_group("input_filtering_params")    
    input_filtering_params.add_argument('--predict_chroms',nargs="*",default=None)
    input_filtering_params.add_argument("--genome",default=None)
    input_filtering_params.add_argument("--fold",type=int,default=None)
    input_filtering_params.add_argument('--center_on_summit',default=False,action='store_true',help="if this is set to true, the peak will be centered at the summit (must be last entry in bed file or hammock) and expanded args.flank to the left and right")
    input_filtering_params.add_argument("--tasks",nargs="*",default=None)
    
    output_params=parser.add_argument_group("output_params")
    output_params.add_argument('--predictions_and_labels_hdf5',help='name of hdf5 to save predictions',default=None)
    calibration_params=parser.add_argument_group("calibration_params")
    calibration_params.add_argument("--calibrate_classification",action="store_true",default=False)
    calibration_params.add_argument("--calibrate_regression",action="store_true",default=False)        
    
    weight_params=parser.add_argument_group("weight_params")
    weight_params.add_argument('--w1',nargs="*",type=float)
    weight_params.add_argument('--w0',nargs="*",type=float)
    weight_params.add_argument("--w1_w0_file",default=None)


    model_params=parser.add_argument_group("model_params")
    model_params.add_argument('--load_model_hdf5',help='hdf5 file that stores the model')
    model_params.add_argument('--weights',help='weights file for the model')
    model_params.add_argument('--yaml',help='yaml file for the model')
    model_params.add_argument('--json',help='json file for the model')
    model_params.add_argument('--functional',default=False,help='use this flag if your model is a functional model',action="store_true")
    model_params.add_argument('--squeeze_input_for_gru',action='store_true')
    model_params.add_argument("--expand_dims",default=False,action='store_true')
    model_params.add_argument("--num_inputs",type=int)
    model_params.add_argument("--num_outputs",type=int)
    model_params.add_argument("--num_gpus",type=int,default=1)
    
    parallelization_params=parser.add_argument_group("parallelization")
    parallelization_params.add_argument("--threads",type=int,default=1)
    parallelization_params.add_argument("--max_queue_size",type=int,default=100)

    snp_params=parser.add_argument_group("snp_params")
    snp_params.add_argument('--background_freqs',default=None)
    snp_params.add_argument('--flank',default=500,type=int)
    snp_params.add_argument('--mask',default=10,type=int)
    snp_params.add_argument('--ref_col',type=int,default=None)
    snp_params.add_argument('--alt_col',type=int,default=None)

    parser.add_argument('--batch_size',type=int,help='batch size to use to make model predictions',default=50)
    return parser.parse_args()

def get_out_predictions_prefix(args):
    if args.predictions_and_labels_hdf5.startswith('s3://'):
        #use a local version of the file and upload to s3 when finished
        bucket,filename=s3_string_parse(args.predictions_and_labels_hdf5)
        out_predictions_prefix=filename.split('/')[-1]+".predictions"
    else: 
        out_predictions_prefix=args.predictions_and_labels_hdf5+".predictions"
    if args.calibrate_classification is True:
        out_predictions_prefix=out_predictions_prefix+".logits"
    elif args.calibrate_regression is True:
        out_predictions_prefix=out_predictions_prefix+".preacts"
    return out_predictions_prefix 

def get_out_calibrated_prefix(args):
    if args.predictions_and_labels_hdf5.startswith('s3://'):
        #use a local version of the file and upload to s3 when finished
        bucket,filename=s3_string_parse(args.predictions_and_labels_hdf5)
        out_predictions_prefix=filename.split('/')[-1]+".predictions.calibrated"
    else: 
        out_predictions_prefix=args.predictions_and_labels_hdf5+".predictions.calibrated"
    return out_predictions_prefix 
    

def get_out_labels_prefix(args):
    if args.predictions_and_labels_hdf5.startswith('s3://'):
        #use a local version of the file and upload to s3 when finished
        bucket,filename=s3_string_parse(args.predictions_and_labels_hdf5)
        out_labels_prefix=filename.split('/')[-1]+".labels"
    else: 
        out_labels_prefix=args.predictions_and_labels_hdf5+".labels" 
    return out_labels_prefix

def write_predictions(args):
    '''
    separate predictions file for each output/task combination 
    '''
    try:
        out_predictions_prefix=get_out_predictions_prefix(args)
        first=True
        while True:
            pred_df=pred_queue.get()
            if type(pred_df) == str: 
                if pred_df=="FINISHED":
                    return
            if first is True:
                mode='w'
                first=False
                append=False
            else:
                mode='a'
                append=True
            for cur_output_index in range(len(pred_df)):
                #get cur_pred_df for current output
                cur_pred_df=pred_df[cur_output_index]
                if args.tasks is not None:
                    task_names={}
                    for i in range(len(args.tasks)):
                        task_names[i]=args.tasks[i]
                    cur_pred_df=cur_pred_df.rename(columns=task_names)
                cur_out_f='.'.join([out_predictions_prefix,str(cur_output_index)])
                cur_pred_df.to_hdf(cur_out_f,key="data",mode=mode,append=append,format="table",min_itemsize={'CHR':30})
                
    except KeyboardInterrupt:
        #shutdown the pool
        # Kill remaining child processes
        kill_child_processes(os.getpid())
        raise 
    except Exception as e:
        print(e)
        #shutdown the pool
        # Kill remaining child processes
        kill_child_processes(os.getpid())
        raise e

def write_labels(args):
    '''
    separate label file for each output/task combination
    '''
    out_labels_prefix=get_out_labels_prefix(args)
    try:
        first=True
        while True:
            label_df=label_queue.get()
            if type(label_df)==str:
                if label_df=="FINISHED":
                    return
            if first is True:
                mode='w'
                first=False
                append=False
            else:
                mode='a'
                append=True
            for cur_output_index in range(len(label_df)):
                cur_label_df=label_df[cur_output_index]
                if args.tasks is not None:
                    task_names={}
                    for i in range(len(args.tasks)):
                        task_names[i]=args.tasks[i]
                    cur_label_df=cur_label_df.rename(columns=task_names)

                cur_out_f='.'.join([out_labels_prefix,str(cur_output_index)])
                cur_label_df.to_hdf(cur_out_f,key="data",mode=mode,append=append,format="table",min_itemsize={'CHR':30})
                
    except KeyboardInterrupt:
        #shutdown the pool
        # Kill remaining child processes
        kill_child_processes(os.getpid())
        raise 
    except Exception as e:
        print(e)
        #shutdown the pool
        # Kill remaining child processes
        kill_child_processes(os.getpid())
        raise e
    return

def init_worker():
    signal.signal(signal.SIGINT, signal.SIG_IGN)

def kill_child_processes(parent_pid, sig=signal.SIGTERM):
    try:
        parent = mp.Process(parent_pid)
    except psutil.NoSuchProcess:
        return
    children = parent.children(recursive=True)
    for process in children:
        process.send_signal(sig)
        


def get_batch_wrapper(idx):
    X,y,coords=test_generator[idx]
    if type(y) is not list:
        y=[y]
    try:
        y=[i.squeeze(axis=-1) for i in y]
    except:
        pass
    if type(X) is not list:
        X=[X]
    
    #represent coords w/ string, MultiIndex not supported in table append mode
    #set the column names for the MultiIndex
    coords=pd.MultiIndex.from_tuples(coords,names=['CHR','START','END'])
    y=[pd.DataFrame(i,index=coords) for i in y]
    return [X,y,coords,idx]


def get_hdf5_predict_generator(args):
    global test_generator
    test_chroms=get_chroms(args,split='test')
    test_generator=DataGenerator(index_path=args.index_data_path,
                                 input_path=args.input_data_path,
                                 output_path=args.output_data_path,
                                 index_tasks=args.index_tasks,
                                 num_inputs=args.num_inputs,
                                 num_outputs=args.num_outputs,
                                 ref_fasta=args.ref_fasta,
                                 batch_size=args.batch_size,
                                 add_revcomp=False,
                                 chroms_to_use=test_chroms,
                                 expand_dims=args.expand_dims,
                                 tasks=args.tasks,
                                 shuffle=False,
                                 return_coords=True)
    return test_generator
def get_variant_predict_generator(args):
    global test_generator
    test_chroms=get_chroms(args,split='test')
    test_generator=SNPGenerator(args.allele_col,
                                args.flank,
                                index_path=args.index_data_path,
                                input_path=args.input_data_path,
                                output_path=args.output_data_path,
                                index_tasks=args.index_tasks,
                                num_inputs=args.num_inputs,
                                num_outputs=args.num_outputs,
                                ref_fasta=args.ref_fasta,
                                allele_col=args.ref_col,
                                batch_size=args.batch_size,
                                add_revcomp=False,
                                chroms_to_use=test_chroms,
                                expand_dims=args.expand_dims,
                                tasks=args.tasks,
                                shuffle=False,
                                return_coords=True)
    return test_generator

def get_generator(args):
    if args.variant_bed is not None:
        return get_variant_predict_generator(args)
    else:
        return get_hdf5_predict_generator(args)

def predict_on_batch_wrapper(args,model,test_generator):
    num_batches=len(test_generator)
    processed=0
    try:
        with mp.Pool(processes=args.threads,initializer=init_worker) as pool: 
            while (processed < num_batches):
                idset=range(processed,min([num_batches,processed+args.max_queue_size]))
                for result in pool.imap_unordered(get_batch_wrapper,idset):
                    X=result[0]
                    y=result[1]
                    coords=result[2]
                    idx=result[3]
                    processed+=1
                    if processed%10==0:
                        print(str(processed)+"/"+str(num_batches))
                    #get the model predictions            
                    preds=model.predict_on_batch(X)
                    if type(preds) is not list:
                        preds=[preds]
                    try:
                        preds=[i.squeeze(axis=-1) for i in preds]
                    except:
                        pass 
                    preds_dfs=[pd.DataFrame(cur_pred,index=coords) for cur_pred in preds]
                    label_queue.put(y)
                    pred_queue.put(preds_dfs)
                    
    except KeyboardInterrupt:
        #shutdown the pool
        pool.terminate()
        pool.join() 
        # Kill remaining child processes
        kill_child_processes(os.getpid())
        raise 
    except Exception as e:
        print(e)
        #shutdown the pool
        pool.terminate()
        pool.join()
        # Kill remaining child processes
        kill_child_processes(os.getpid())
        raise e
    print("finished with tiledb predictions!")
    label_queue.put("FINISHED")
    label_queue.close() 
    pred_queue.put("FINISHED")
    pred_queue.close() 
    return


def get_model_layer_functor(model,target_layer_idx):
    from keras import backend as K
    inp=model.input
    outputs=model.layers[target_layer_idx].output
    functor=K.function([inp], [outputs])
    return functor 

def get_layer_outputs(functor,X):
    return functor([X])

def predict(args):
    if type(args)==type({}):
        args=args_object_from_args_dict(args) 
    global pred_queue
    global label_queue
    
    pred_queue=mp.Queue()
    label_queue=mp.Queue()
    
    label_writer=mp.Process(target=write_predictions,args=([args]))
    pred_writer=mp.Process(target=write_labels,args=([args]))
    label_writer.start()
    pred_writer.start() 


    #get the generator
    test_generator=get_generator(args) 
    
    #get the model
    #if calibration is to be done, get the preactivation model 
    model=get_model(args)
    perform_calibration=args.calibrate_classification or args.calibrate_regression
    if perform_calibration==True:
        if args.calibrate_classification==True:
            print("getting logits")
            model=Model(inputs=model.input,
                               outputs=model.layers[-2].output)
        elif args.calibrate_regression==True:
            print("getting pre-relu outputs (preacts)")
            model=Model(inputs=model.input,
                        outputs=model.layers[-1].output)
            
    #call the predict_on_batch_wrapper
    predict_on_batch_wrapper(args,model,test_generator)

    #drain the queue
    try:
        while not label_queue.empty():
            print("draining the label Queue")
            time.sleep(2)
    except Exception as e:
        print(e)
    try:
        while not pred_queue.empty():
            print("draining the prediction Queue")
            time.sleep(2)
    except Exception as e:
        print(e)
    
    print("joining label writer") 
    label_writer.join()
    print("joining prediction writer") 
    pred_writer.join()

    #sync files to s3 if needed
    if args.predictions_and_labels_hdf5.startswith('s3://'):
        #use a local version of the file and upload to s3 when finished
        bucket,filename=s3_string_parse(args.predictions_and_labels_hdf5)
        out_predictions_prefix=filename.split('/')[-1]+".predictions"
        out_labels_prefix=filename.split('/')[-1]+".predictions"
        #upload outputs for all tasks
        import glob
        to_upload=[]
        for f in glob.glob(out_predictions_prefix+"*"):
            s3_path='/'.join(args.predictions_and_labels_hdf5.split('/')[0:-1])+"/"+f
            upload_s3_file(s3_path,f)
        for f in glob.glob(out_labels_prefix+"*"):
            s3_path='/'.join(args.predictions_and_labels_hdf5.split('/')[0:-1])+"/"+f
            upload_s3_file(s3_path,f)
        
    #perform calibration, if specified
    if perform_calibration is True:
        print("calibrating")
        out_predictions_prefix=get_out_predictions_prefix(args)
        out_labels_prefix=get_out_labels_prefix(args)
        out_calibrated_prefix=get_out_calibrated_prefix(args)
        
        for output_index in range(args.num_outputs):
            preacts_fname='.'.join([out_predictions_prefix,str(output_index)])
            labels_fname='.'.join([out_labels_prefix,str(output_index)])
            calibrated_fname='.'.join([out_calibrated_prefix,str(output_index)])
            #model has already been transformed to a preact model above
            calibrate(preacts_fname,
                      labels_fname,
                      model,
                      calibrated_fname,
                      calibrate_regression=args.calibrate_regression,
                      calibrate_classification=args.calibrate_classification,
                      get_model_preacts=False)
    #clean up any s3 artifacts:
    run_cleanup()
    
def main():
    args=parse_args()
    predict(args)


if __name__=="__main__":
    main()
    



import argparse
def parse_args():
    parser=argparse.ArgumentParser()
    parser.add_argument("--file_name_prefix")
    parser.add_argument("--file_name_suffix")
    parser.add_argument("--outf")
    parser.add_argument("--nsplits",type=int,default=10)
    return parser.parse_args()

def main():
    args=parse_args()
    aggregate_dict={}
    tasks=None
    for split in range(args.nsplits):
        try:
            aggregate_dict[split]=dict()
            fname=args.file_name_prefix+str(split)+args.file_name_suffix
            print(fname)
            data=open(fname,'r').read().strip().split('\n')
            if tasks is None:
                header=data[0].split('\t')
                tasks=header
                print(str(tasks))
            for line in data[1::]:
                tokens=line.split('\t')
                metric=tokens[0]
                aggregate_dict[split][metric]={} 
                for i in range(len(tasks)):
                    cur_task=tasks[i] 
                    cur_val=round(float(tokens[i+1].split('(')[-1].split(')')[0].split(',')[0].split('=')[-1]),3)
                    #print(cur_val)
                    aggregate_dict[split][metric][cur_task]=cur_val
        except:
            print("skipping split:"+str(split))
    print(aggregate_dict)
    outf=open(args.outf,'w')
    outf.write('Split\tMetrics\t'+'\t'.join(tasks)+'\n')
    for split in range(args.nsplits):
        for metric in aggregate_dict[split]:
            try:
                outf.write(str(split)+'\t'+str(metric))
                for task in tasks:
                    outf.write('\t'+str(aggregate_dict[split][metric][task]))
                outf.write('\n')
            except:
                continue 
                
if __name__=="__main__":
    main()
    
#prefix="gecco.200bpregression.with.gc.performance.fold"
#suffix=".tsv"



import argparse
import pdb 
def args_object_from_args_dict(args_dict):
    #create an argparse.Namespace from the dictionary of inputs
    args_object=argparse.Namespace()
    #set the defaults for training/prediction/interpretation/cross-validation
    #training
    vars(args_object)['data_path']=None
    vars(args_object)['tiledb_tasks_file']=None
    vars(args_object)['chrom_sizes']=None
    vars(args_object)['label_aggregation']=None
    vars(args_object)['nonzero_bin_path']=None
    vars(args_object)['universal_negative_path']=None 
    vars(args_object)['multigpu']=False
    vars(args_object)['batch_size']=1000
    vars(args_object)['num_train']=700000
    vars(args_object)['num_valid']=150000
    vars(args_object)['train_chroms']=None
    vars(args_object)['validation_chroms']=None 
    vars(args_object)['init_weights']=None
    vars(args_object)['w1_w0_file']=None
    vars(args_object)['squeeze_input_for_gru']=False 
    vars(args_object)['from_checkpoint_weights']=None
    vars(args_object)['from_checkpoint_arch']=None
    vars(args_object)['vcf_file']=None
    vars(args_object)['epochs']=40
    vars(args_object)['patience']=3
    vars(args_object)['patience_lr']=2
    vars(args_object)['architecture_spec']="basset_architecture_multitask"
    vars(args_object)['architecture_from_file']=None
    vars(args_object)['tensorborad_logdir']='logs'
    vars(args_object)['seed']=1234
    vars(args_object)['train_path']=None
    vars(args_object)['valid_path']=None
    vars(args_object)['upsample_thresh']=0
    vars(args_object)['train_upsample']=None
    vars(args_object)['valid_upsample']=None
    vars(args_object)['threads']=1
    vars(args_object)['max_queue_size']=100
    vars(args_object)['save_weights']=None
    vars(args_object)['w1']=None
    vars(args_object)['w0']=None
    vars(args_object)['yaml']=None
    vars(args_object)['json']=None
    vars(args_object)['expand_dims']=False
    vars(args_object)['tasks']=None
    vars(args_object)['shuffle_epoch_start']=False
    vars(args_object)['shuffle_epoch_end']=False
    vars(args_object)['revcomp']=False
    
    #prediction
    vars(args_object)['predict_chroms']=None
    vars(args_object)['prediction_pickle']=None
    vars(args_object)['performance_metrics_classification_file']=None
    vars(args_object)['performance_metrics_regression_file']=None
    vars(args_object)['predictions_pickle_to_load']=None
    vars(args_object)['sequential']=False
    vars(args_object)['background_freqs']=None
    vars(args_object)['variant_bed']=None
    vars(args_object)['mask']=10
    vars(args_object)['center_on_summit']=False
    vars(args_object)['calibrate_classification']=False
    vars(args_object)['calibrate_regression']=False
    vars(args_object)['tiledb_stride']=1
    vars(args_object)['label_subset_attribute']=None
    vars(args_object)['label_thresh']=None

    #cross-validation
    vars(args_object)['splits']=None

    #interpret
    vars(args_object)['method']='deeplift'
    vars(args_object)['deeplift_ref']='shuffled_ref' 
    vars(args_object)['background_freqs']=None
    vars(args_object)['precision_thresh']=0.90
    for key in args_dict:
        vars(args_object)[key]=args_dict[key]
    args=args_object
    return args

    
    



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import pandas as pd
from .custom_losses import *
from .metrics import *
from keras.models import load_model
from keras.models import Model
from keras.utils.generic_utils import get_custom_objects
from abstention.calibration import PlattScaling, IsotonicRegression 
import argparse
def parse_args():
    parser=argparse.ArgumentParser(description="calibration of model preacts/logits")
    parser.add_argument("--preacts",help="preact/logit hdf5 file generated by kerasAC_predict")
    parser.add_argument("--labels",help="hdf5 file generated by kerasAC_predict")
    parser.add_argument("--model",help="hdf5 file generated by kerasAC_train")
    parser.add_argument("--outf",help="name of output hdf5 file")
    parser.add_argument("--calibrate_regression",action="store_true",default=False)
    parser.add_argument("--calibrate_classification",action="store_true",default=False) 
    return parser.parse_args() 

def load_model_wrapper(model_hdf5_fname):
    custom_objects={"recall":recall,
                    "sensitivity":recall,
                    "specificity":specificity,
                    "fpr":fpr,
                    "fnr":fnr,
                    "precision":precision,
                    "f1":f1,
                    "ambig_binary_crossentropy":ambig_binary_crossentropy,
                    "ambig_mean_absolute_error":ambig_mean_absolute_error,
                    "ambig_mean_squared_error":ambig_mean_squared_error,
                    "MultichannelMultinomialNLL":MultichannelMultinomialNLL}
    get_custom_objects().update(custom_objects)
    model=load_model(model_hdf5_fname)
    return model

def get_preacts(model,calibrate_classification=False, calibrate_regression=False):
    if calibrate_classification==True:
        print("getting logits")
        return Model(inputs=model.input,
                     outputs=model.layers[-2].output)
    elif calibrate_regression==True:
        print("getting pre-relu outputs (preacts)")
        return Model(inputs=model.input,
                    outputs=model.layers[-1].output)


def calibrate(preacts,labels,model,outf,calibrate_regression=False,calibrate_classification=False,get_model_preacts=False):
    assert not ((calibrate_classification==False) and (calibrate_regression==False))
    assert not ((calibrate_classification==True) and (calibrate_regression==True))
    if(type(preacts)!=type(pd.DataFrame)):
        #load the preacts
        preacts=pd.read_hdf(preacts)
        print("loaded preacts from hdf5")
    if(type(labels)!=type(pd.DataFrame)):
        labels=pd.read_hdf(labels)
        print("loaded labels from hdf5") 
    #make sure they are in the same order
    labels=labels.loc[preacts.index]
    print("ordered labels")
    if type(model)==str:
        model=load_model_wrapper(model)
        get_model_preact=True 
    if get_model_preact==True:
        model=get_preacts(model,calibrate_classification=calibrate_classification, calibrate_regression=calibrate_regression)
        
        
    #perform calibration for each task!
    calibrated_predictions=None
    for i in range(preacts.shape[1]):
        #don't calibrate on nan inputs
        nonambiguous_indices=np.argwhere(~np.isnan(labels[i]))
        if calibrate_classification==True:
            calibration_func = PlattScaling()(valid_preacts=preacts[i][nonambiguous_indices],
                                                             valid_labels=labels[i][nonambiguous_indices])
        elif calibrate_regression==True:
            calibration_func=IsotonicRegression()(valid_preacts=preacts[i][nonambiguous_indices].squeeze(),
                                                  valid_labels=labels[i][nonambiguous_indices].squeeze())

        calibrated_predictions_task=calibration_func(preacts[i].values)
        if calibrated_predictions is None:
            calibrated_predictions=np.expand_dims(calibrated_predictions_task,axis=1)
        else:
            calibrated_predictions=np.concatenate((calibrated_predictions,np.expand_dims(calibrated_predictions_task,axis=1)),axis=1)
    calibrated_predictions=pd.DataFrame(calibrated_predictions,index=preacts.index)
    calibrated_predictions.to_hdf(outf,key="data",mode='w',append=False,format="table",min_itemsize={'CHR':30})

def main():     
    args=parse_args()
    calibrate(args.preacts,args.labels,args.model,args.outf,calibrate_regression=args.calibrate_regression, calibrate_classification=args.calibrate_classification,get_model_preacts=True)

    
if __name__=="__main__":
    main()



import tensorflow as tf


pseudocount=0.01


def contingency_table(y, z):
    """Note:  if y and z are not rounded to 0 or 1, they are ignored
    """
    y = tf.cast(tf.round(y), tf.floatx())
    z = tf.cast(tf.round(z), tf.floatx())
    
    def count_matches(y, z):
        return tf.sum(tf.cast(y, tf.floatx()) * tf.cast(z, tf.floatx()))
    
    ones = tf.ones_like(y)
    zeros = tf.zeros_like(y)
    y_ones = tf.equal(y, ones)
    y_zeros = tf.equal(y, zeros)
    z_ones = tf.equal(z, ones)
    z_zeros = tf.equal(z, zeros)
    
    tp = count_matches(y_ones, z_ones)
    tn = count_matches(y_zeros, z_zeros)
    fp = count_matches(y_zeros, z_ones)
    fn = count_matches(y_ones, z_zeros)
    return (tp, tn, fp, fn)

def recall(y, z):
    """True positive rate `tp / (tp + fn)`
    """
    tp, tn, fp, fn = contingency_table(y, z)
    return tp / (tp + fn+pseudocount)

def tpr(y, z):
    """
    True positive rate `tp / (tp + fn)`
    """
    tp, tn, fp, fn = contingency_table(y, z)
    return tp / (tp + fn)

def tnr(y, z):
    """
    True negative rate `tn / (tn + fp)`
    """
    tp, tn, fp, fn = contingency_table(y, z)
    return tn / (tn + fp)

def specificity(y, z):
    """True negative rate `tn / (tn + fp)`
    """
    tp, tn, fp, fn = contingency_table(y, z)
    return tn / (tn + fp+pseudocount)


def fpr(y, z):
    """False positive rate `fp / (fp + tn)`
    """
    tp, tn, fp, fn = contingency_table(y, z)
    return fp / (fp + tn+pseudocount)


def fnr(y, z):
    """False negative rate `fn / (fn + tp)`
    """
    tp, tn, fp, fn = contingency_table(y, z)
    return fn / (fn + tp+pseudocount)


def precision(y, z):
    """Precision `tp / (tp + fp)`
    """
    tp, tn, fp, fn = contingency_table(y, z)
    return tp / (tp + fp+pseudocount)


def fdr(y, z):
    """False discovery rate `fp / (tp + fp)`
    """
    tp, tn, fp, fn = contingency_table(y, z)
    return fp / (tp + fp+pseudocount)


def f1(y, z):
    """F1 score: `2 * (p * r) / (p + r)`, where p=precision and r=recall.
    """
    _recall = recall(y, z)
    _prec = precision(y, z)
    return 2 * (_prec * _recall) / (_prec + _recall+pseudocount)




import numpy as np ;
from kerasAC.metrics import *
from keras.constraints import max_norm
from kerasAC.custom_losses import get_weighted_binary_crossentropy, get_ambig_binary_crossentropy, get_ambig_mean_squared_error
from kerasAC.metrics import recall, specificity, fpr, fnr, precision, f1
import keras
from keras.models import Sequential
from keras.layers.core import Dropout, Reshape, Dense, Activation, Flatten
from keras.layers.convolutional import Conv2D, MaxPooling2D
from keras.optimizers import Adadelta, SGD, RMSprop;
import keras.losses;
from keras.constraints import maxnorm;
from keras.layers.normalization import BatchNormalization
from keras.regularizers import l1, l2    
from keras import backend as K
K.set_image_data_format('channels_last')
print(K.image_data_format())

def getModelGivenModelOptionsAndWeightInits(args):
    #read in the args
    ntasks=args.ntasks
    seed=args.seed
    
    np.random.seed(seed)
    model=Sequential()
    model.add(Conv2D(filters=50,kernel_size=(1,15),padding="same", kernel_constraint=max_norm(7.0,axis=-1),input_shape=(1,1000,4)))
    model.add(BatchNormalization(axis=-1))
    model.add(Activation('relu'))   
    model.add(Conv2D(filters=50,kernel_size=(1,15),padding="same"))
    model.add(BatchNormalization(axis=-1))
    model.add(Activation('relu'))    
    model.add(Conv2D(filters=50,kernel_size=(1,13),padding="same"))
    model.add(BatchNormalization(axis=-1))
    model.add(Activation('relu'))    
    model.add(MaxPooling2D(pool_size=(1,40)))    
    model.add(Flatten())
    model.add(Dense(50))
    model.add(BatchNormalization(axis=-1))
    model.add(Activation('relu'))
    model.add(Dropout(0.2))    
    model.add(Dense(ntasks))
    adam = keras.optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
    print("compiling!")
    loss=get_ambig_mean_squared_error() 
    model.compile(optimizer=adam,
                  loss=loss,
                  metrics=[recall, specificity, fpr, fnr, precision, f1])
    return model



from setuptools import setup,find_packages

config = {
    'include_package_data': True,
    'description': 'keras Accessibility Models (kerasAC)',
    'download_url': 'https://github.com/kundajelab/kerasAC',
    'version': '2.5.3',
    'packages': ['kerasAC'],
    'setup_requires': [],
    'install_requires': ['tensorflow-gpu>=2.3','pysam', 'tiledb>=0.5.2','psutil','tables','numpy>=1.9', 'keras>=2.4', 'h5py', 'pandas','pybigwig','deeplift','abstention','boto3'],
    'scripts': [],
    'entry_points': {'console_scripts': ['kerasAC_train = kerasAC.train:main',
                                         'kerasAC_predict_hdf5=kerasAC.predict_hdf5:main',
                                         'kerasAC_predict_tdb=kerasAC.predict_tiledb:main',
                                         'kerasAC_calibrate=kerasAC.calibrate:main',
                                         'kerasAC_curves=kerasAC.curves:main',
                                         'kerasAC_score=kerasAC.performance_metrics.performance_metrics:main',
                                         'kerasAC_score_bpnet=kerasAC.performance_metrics.bpnet_performance_metrics:main',
                                         'kerasAC_score_bpnet_legacy=kerasAC.performance_metrics.bpnet_performance_metrics_legacy:main',
                                         'kerasAC_interpret=kerasAC.interpret:main',
                                         'kerasAC_bpnet_shap_wrapper=kerasAC.interpret.bpnet_shap_wrapper:main',
                                         'kerasAC_plot_interpretation=kerasAC.plot_interpretation:main',
                                         'kerasAC_cross_validate=kerasAC.cross_validate:main',
                                         'kerasAC_loss_weights_bpnet=kerasAC.helpers.get_loss_weights_for_bpnet:main',
                                         'kerasAC_bigwigs_from_io=kerasAC.helpers.generate_bigwigs_from_bpnet_io:main',
                                         'kerasAC_aggregate_summaries=kerasAC.helpers.aggregate_summaries:main']},
    'name': 'kerasAC'
}

if __name__== '__main__':
    setup(**config)



import numpy as np
from kerasAC.metrics import *
import keras;
import pdb;
from keras.models import Model
from keras.layers import Input, Bidirectional
from keras.layers.merge import Add, Multiply, Dot
from keras.layers.recurrent import GRU
from keras.layers.core import Dropout, Dense, Activation, Reshape
from keras.optimizers import Adam;
import keras.losses;
import keras.activations
from keras.layers.normalization import BatchNormalization
from keras import backend as K

from keras.activations import softmax

def softMaxAxis1(x):
    return softmax(x,axis=1)
    

#note: functional model
#this is single-headed battnet (i.e. one attention head only) 
def getModelGivenModelOptionsAndWeightInits(args):
    #read in the arguments
    w0=args.w0
    w1=args.w1
    init_weights=args.init_weights
    ntasks=args.ntasks
    seed=args.seed
    
    np.random.seed(seed)
    K.set_image_data_format('channels_last')
    print(K.image_data_format())
    inputs=Input(shape=(1000,4))    
    gru_output=Bidirectional(GRU(512,dropout=0.3,input_dim=(1000,4),implementation=2,return_sequences=True))(inputs)
    
    
    #first output is attention
    attn1=Dense(1024)(gru_output)
    attn1=BatchNormalization(axis=-1)(attn1)
    attn1=Activation("relu")(attn1)
    attn1=Dropout(0.3)(attn1)

    attn1=Dense(1)(attn1)
    attn1=Activation("relu")(attn1)
    
    attn1_output=Activation(softMaxAxis1)(attn1)
    #add in fc 3 - 5
    added=Dot(axes=1)([gru_output,attn1_output])
    added=Reshape((1024,))(added)
    x=Dense(512)(added)
    x=BatchNormalization(axis=-1)(x)
    x=Activation("relu")(x)
    x=Dropout(0.3)(x)

    x=Dense(512)(x)
    x=BatchNormalization(axis=-1)(x)
    x=Activation("relu")(x)
    intermediate_out=Dropout(0.3)(x)

    #the output layer    
    outputs=Dense(ntasks,activation="sigmoid")(intermediate_out)
    
    model=Model(inputs=inputs,outputs=outputs)
    print(model.summary())
    adam = keras.optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-08)
    print("compiling!")
    if w0!=None:
        import kerasAC.custom_losses
        loss=kerasAC.custom_losses.get_weighted_binary_crossentropy(w0_weights=w0,w1_weights=w1)
    else:
        loss="binary_crossentropy"
    model.compile(optimizer=adam,loss=loss,metrics=[positive_accuracy,negative_accuracy,precision,recall])
    return model



import importlib
import imp

from .s3_sync import *
from .custom_losses import *
from .metrics import *

def get_w1_w0_training(args,train_generator):
    w1=args.w1
    w0=args.w0
    if (args.w1_w0_file is not None) and (args.w1_w0_file.startswith('s3://')):
        w1_w0_file=download_s3_file(args.w1_w0_file)
    else:
        w1_w0_file=args.w1_w0_file
    if (args.weighted==True and (w1==None or w0==None) ):
        if args.w1_w0_file==None:
            w1=train_generator.w1
            w0=train_generator.w0        
            assert args.save_w1_w0 !=None
            with open(args.save_w1_w0, 'w') as weight_file:
                for i in range(len(w1)):
                    weight_file.write(str(w1[i])+'\t'+str(w0[i])+'\n')
        else:
            w1_w0=np.loadtxt(args.w1_w0_file)
            w1=list(w1_w0[:,0])
            w0=list(w1_w0[:,1]) 
        print("got weights!")
    return w1,w0

def get_w1_w0_prediction(args):
    w1=None
    w0=None
    if args.w1_w0_file is not None:
        if args.w1_w0_file.startswith('s3://'):
            w1_w0_file=download_s3_file(args.w1_w0_file)
        else:
            w1_w0_file=args.w1_w0_file
        w1_w0=np.loadtxt(w1_w0_file)
        w1=w1_w0[:,0]
        w0=w1_w0[:,1]
    if args.w1!=None:
        w1=args.w1
    if args.w0!=None:
        w0=args.w0 
    return w1,w0

def get_model(args):
    from tensorflow.keras.utils import get_custom_objects
    custom_objects={"recall":recall,
                    "sensitivity":recall,
                    "specificity":specificity,
                    "fpr":fpr,
                    "fnr":fnr,
                    "precision":precision,
                    "f1":f1,
                    "ambig_binary_crossentropy":ambig_binary_crossentropy,
                    "ambig_mean_absolute_error":ambig_mean_absolute_error,
                    "ambig_mean_squared_error":ambig_mean_squared_error,
                    "MultichannelMultinomialNLL":MultichannelMultinomialNLL,        
                    "MultichannelMultinomialMSE":MultichannelMultinomialMSE}        
    w1,w0=get_w1_w0_prediction(args)
    if type(w1) in [np.ndarray, list]: 
        loss_function=get_weighted_binary_crossentropy(w0,w1)
        custom_objects["weighted_binary_crossentropy"]=loss_function
    get_custom_objects().update(custom_objects)
    
    if args.yaml!=None:
        from tensorflow.keras.models import model_from_yaml
        #load the model architecture from yaml
        if args.yaml.startswith('s3://'):
            yaml_string=read_s3_file_contents(args.yaml)
        else: 
            yaml_string=open(args.yaml,'r').read()
        model=model_from_yaml(yaml_string)#,custom_objects=custom_objects) 
    elif args.json!=None:
        from tensorflow.keras.models import model_from_json
        #load the model architecture from json
        if args.json.startswith('s3://'):
            json_string=read_s3_file_contents(args.json)
        else: 
            json_string=open(args.json,'r').read()
        model=model_from_json(json_string)#,custom_objects=custom_objects)
    elif args.load_model_hdf5!=None: 
        #load from the hdf5
        from tensorflow.keras.models import load_model
        if args.load_model_hdf5.startswith("s3://"):
            model_hdf5=download_s3_file(args.load_model_hdf5)
        else:
            model_hdf5=args.load_model_hdf5
        model=load_model(model_hdf5)
    else:
        #initialize model from user-supplied architecture
        try:
            if (args.architecture_from_file!=None):
                architecture_module=imp.load_source('',args.architecture_from_file)
            else:
                architecture_module=importlib.import_module('kerasAC.architectures.'+args.architecture_spec)
        except:
            raise Exception("could not import requested architecture, is it installed in kerasAC/kerasAC/architectures? Is the file with the requested architecture specified correctly?")
        model=architecture_module.getModelGivenModelOptionsAndWeightInits(args)
    if args.num_gpus >1:
        try:
            model=multi_gpu_model(model,gpus=args.num_gpus)
            print("Training/predicting on" +str(args.num_gpus)+" GPU's. Set args.multi_gpu = False to avoid this") 
        except:
            print("failed to instantiate multi-gpu model, defaulting to single-gpu model")
    model=load_model_weights(args.weights,model)
    print("prepared the model")
    model.summary()
    return model

def load_model_weights(weight_file,model):
    if weight_file is None:
        #nothing to do 
        return model 
    #sync the model locally if it's on AWS
    if weight_file.startswith("s3://"):
        s3_model_weights=download_s3_file(weight_file)
    else:
        s3_model_weights=weight_file
    import  h5py
    try:
        model.load_weights(s3_model_weights)#,by_name=True)
    except:
        with h5py.File(s3_model_weights,'r') as file:
            weight_file=file['model_1']
            for layer in model.layers:
                try:
                    layer_weights=weight_file[layer.name]
                except:
                    print('no weights files saved for layer:'+str(layer.name))
                    continue
                try:
                    weights = []
                    # Extract weights
                    for term in layer_weights:
                        if isinstance(layer_weights[term], h5py.Dataset):
                            # Convert weights to numpy array and prepend to list
                            weights.insert(0, np.array(layer_weights[term]))        
                    # Load weights to model
                    layer.set_weights(weights)
                    print("loaded weights for layer:"+str(layer.name))
                except Exception as e:
                    print("Error: Could not load weights for layer:"+str(layer.name))
                    raise e
    return model 
